/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pws.h"

const char *STR_ERROR[] = 
{
    "Success.",
    "Not a valid PWS file.",
    "Wrong Key.",
    "PWS file is either corrupt or truncated.",
    "Write Error.",
    "libgcrypt errror."
};

/**
 * Initializes the pws_context_t object,
 * Must be called before using any other decrypting functions.
 *
 * @param ctx - The pws_context_t context object containing the context that should be initialized.
 * @param p   - The passphrase.
 * @param enc_only - If set to 1, only the pwd hash will be hold in memory, the pwd itself
 *                   will be cleaned immediately. Only valid if this context will be used
 *                   only for encrypting purposes. If this context is used for decryption also,
 *                   dec_only must be set to 0.
 */
void pws_init_context (pws_context_t *ctx, char *p, int enc_only)
{
    pws3_init_ctx (&ctx->pws3, p, enc_only, &ctx->error);
    pws2_init_ctx (&ctx->pws2, p, enc_only, &ctx->error);

    ctx->error.error = 0;
    ctx->error.str_error = NULL;
}

/**
 * Frees/trashes all ressources used by the pws_context_t object.
 *
 * @param ctx - The pws_context_t context object containing the context that should be freed.
 */
void pws_release_context (pws_context_t *ctx)
{
    pws3_release_ctx (&ctx->pws3);
    pws2_release_ctx (&ctx->pws2);

    free (ctx->error.str_error);
    ctx->error.error = 0;
}

/**
 * Sets all members of hdr to NULL.
 *
 * @param hdr - The pws2_hdr_t object.
 */
void pws2_init_hdr (pws2_hdr_t *hdr)
{
    hdr->notes        = NULL;
    hdr->notes_len    = 0;
    hdr->password     = NULL;
    hdr->password_len = 0;
    hdr->name         = NULL;
    hdr->name_len     = 0;
}

/**
 * Frees/trashes all memory used by hdr.
 *
 * @param hdr - The pws2_hdr_t object to be freed.
 */
void pws2_release_hdr (pws2_hdr_t *hdr)
{
    free (hdr->notes);
    free (hdr->password);
    free (hdr->name);

    pws2_init_hdr (hdr);
}

/**
 * Sets all members of hdr to NULL.
 *
 * @param hdr - The pws3_hdr_t object.
 */
void pws3_init_hdr (pws3_hdr_t *hdr)
{
    hdr->version           = NULL;
    hdr->version_len       = 0;
    hdr->uuid              = NULL;
    hdr->uuid_len          = 0;
    hdr->tree_status       = NULL;
    hdr->tree_status_len   = 0;
    hdr->last_app          = NULL;
    hdr->last_app_len      = 0;
    hdr->last_uid          = NULL;
    hdr->last_uid_len      = 0;
    hdr->last_host         = NULL;
    hdr->last_host_len     = 0;
    hdr->db_name           = NULL;
    hdr->db_name_len       = 0;
    hdr->db_desc           = NULL;
    hdr->db_desc_len       = 0;
    hdr->unknown           = NULL;
    hdr->unknown_len       = 0;
    hdr->last_modified     = 0;
}

/**
 * Frees/trashes all memory used by hdr.
 *
 * @param hdr - The pws3_hdr_t object to be freed.
 */
void pws3_release_hdr (pws3_hdr_t *hdr)
{
    free (hdr->version);
    free (hdr->uuid);
    free (hdr->tree_status);
    free (hdr->last_app);
    free (hdr->last_uid);
    free (hdr->last_host);
    free (hdr->db_name);
    free (hdr->db_desc);
    free (hdr->unknown);

    pws3_init_hdr (hdr);
}

/**
 * Sets all members of pwd_history to NULL.
 *
 * @param pwd_history - The pws_pwd_hst_entry_t object.
 */
void pws_init_pwd_hst_entry (pws_pwd_hst_entry_t *pwd_history)
{
    pwd_history->pwd       = NULL;
    pwd_history->pwd_len   = 0;
    pwd_history->timestamp = 0;
}

/**
 * Frees/trashes all memory used by pwd_history.
 *
 * @param entry - The pws_pwd_hst_entry_t object to be freed.
 */
void pws_release_pwd_hst_entry (pws_pwd_hst_entry_t *pwd_history)
{
    free_secure (pwd_history->pwd);
    pws_init_pwd_hst_entry (pwd_history);
}

/**
 * Sets all members of entry to NULL.
 *
 * @param entry - The pws_entry_t object.
 */
void pws_init_entry (pws_entry_t *entry)
{
    entry->uuid            = NULL;
    entry->uuid_len        = 0;
    entry->group           = NULL;
    entry->group_len       = 0;
    entry->title           = NULL;
    entry->title_len       = 0;
    entry->username        = NULL;
    entry->username_len    = 0;
    entry->notes           = NULL;
    entry->notes_len       = 0;
    entry->pwd             = NULL;
    entry->pwd_len         = 0;
    entry->pwd_policy      = NULL;
    entry->pwd_policy_len  = 0;
    entry->url             = NULL;
    entry->url_len         = 0;
    entry->autotype        = NULL;
    entry->autotype_len    = 0;
    entry->unknown         = NULL;
    entry->unknown_len     = 0;
    entry->ctime           = 0;
    entry->pwd_mtime       = 0;
    entry->mtime           = 0;
    entry->atime           = 0;
    entry->ltime           = 0;
    entry->pwd_history     = NULL;
    entry->pwd_history_len = 0;
    entry->save_pwd_history = -1; 
    entry->def_pwd_max      = -1;
}

/**
 * Frees/trashes all memory used by entry.
 *
 * @param entry - The pws_entry_t object to be freed.
 */
void pws_release_entry (pws_entry_t *entry)
{
    free_secure (entry->uuid);
    free_secure (entry->group);
    free_secure (entry->title);
    free_secure (entry->url);
    free_secure (entry->autotype);
    free_secure (entry->unknown);

    free_secure (entry->username);
    free_secure (entry->notes);
    free_secure (entry->pwd);
    free_secure (entry->pwd_policy);

    entry->uuid_len        = 0;
    entry->group_len       = 0;
    entry->title_len       = 0;
    entry->url_len         = 0;
    entry->autotype_len    = 0;
    entry->username_len    = 0;
    entry->notes_len       = 0;
    entry->pwd_len         = 0;
    entry->save_pwd_history = 0;

    int i; for (i = 0; i < entry->pwd_history_len; i++)
        { pws_release_pwd_hst_entry (&entry->pwd_history[i]); }
    free (entry->pwd_history);

    entry->pwd_policy_len  = 0;
    entry->pwd_history_len = 0;
    
    pws_init_entry (entry);
}


/**
 * Sets all members of prefs to default values.
 *
 * @param prefs - The pws_prefs_t object.
 */
void pws_init_prefs (pws_prefs_t *prefs)
{
    prefs->showpwdefault                = 0;
    prefs->showpasswordintree           = 0;
    prefs->sortascending                = 1;
    prefs->usedefuser                   = 0;
    prefs->saveimmediately              = 1;
    prefs->pwuselowercase               = 1;
    prefs->pwuseuppercase               = 1;
    prefs->pwusedigits                  = 1;
    prefs->pwusesysmbols                = 0;
    prefs->pwusehexdigit                = 0;
    prefs->pweasyvision                 = 0;
    prefs->isutf8                       = 0;
    prefs->displayexpandedaddeditdialog = 1;
    prefs->maintaindatetimestamps       = 0;
    prefs->savepasswordhistory          = 0;
    prefs->shownotesdefault             = 0;
    prefs->showusernameintree           = 1;

    prefs->pwlendefault            = 8;
    prefs->idletimeout             = 5;
    prefs->treedisplaystatusatopen = 0;
    prefs->numpwhistorydefault     = 3;

    prefs->defaultusername              = NULL;
    prefs->defaultusername_len          = 0; 
    prefs->defaultautootypestring       = NULL;
    prefs->defaultautootypestring_len   = 0;

    prefs->unknown     = NULL;
    prefs->unknown_len = 0;
}

/**
 * Frees/trashes all memory used by prefs.
 *
 * @param prefs - The pws_prefs_t object to be freed.
 */
void pws_release_prefs (pws_prefs_t *prefs)
{
    free (prefs->defaultusername);
    free (prefs->defaultautootypestring);
    free (prefs->unknown);

    pws_init_prefs (prefs);
}

/**
 * Sets all members of dat to NULL.
 *
 * @param dat - The pws_data_t object.
 */
void pws_init_data (pws_data_t *dat)
{
    pws2_init_hdr (&dat->hdr2);
    pws3_init_hdr (&dat->hdr3);

    pws_init_prefs(&dat->prefs);

    dat->entries     = NULL;
    dat->num_entries = 0;
}


/**
 * Frees/trashes all memory used by dat.
 *
 * @param dat - The pws_data_t object to be freed.
 */
void pws_release_data (pws_data_t *dat)
{
    pws2_release_hdr (&dat->hdr2);
    pws3_release_hdr (&dat->hdr3);

    pws_release_prefs (&dat->prefs);

    int x;
    for (x = 0; x < dat->num_entries; x++)
    {
        pws_release_entry (dat->entries[x]);
        free (dat->entries[x]);
    }

    free (dat->entries);
    dat->num_entries = 0;
}

/**
 * Parses one string value from the nondef string and stores
 * the value inside prefs.
 *
 * @param prefs - The pws_prefs_t object were all pref values will be stored.
 * @param buf   - The single string value from the nondef string.
 *
 * @returns The start of the next preference inside the nondef String.
 */
unsigned char * pws_parse_pref_string (pws_prefs_t * prefs, char *buf)
{
    char *beg = buf;

    int typ = strtol (buf, &buf, 10);
    buf += 2;

    char * delim = strchr (buf, '"');
    while (delim && *(delim-1) == '\\')
    {
        delim = strchr (delim+1, '"');
    }

    if (!delim)
        { return (NULL); };

    *delim = 0;
    switch (typ)
    {
        case PWS_PREFS_STRING_DEFAULTUSERNAME:
            prefs->defaultusername = malloc ((strlen (buf)+1)*sizeof(char));
            strcpy (prefs->defaultusername, buf);
            prefs->defaultusername_len = strlen (buf);
            break;
        case PWS_PREFS_STRING_DEFAULTAUTOOTYPESTRING:
            prefs->defaultautootypestring = malloc ((strlen (buf)+1)*sizeof(char));
            strcpy (prefs->defaultautootypestring, buf);
            prefs->defaultautootypestring_len = strlen (buf);
            break;
        default:
            if (!prefs->unknown)
            {
                prefs->unknown = strdup ("");
                prefs->unknown_len = 1;
            }
            
            prefs->unknown_len += strlen (beg) + 4;
            prefs->unknown = realloc (prefs->unknown, prefs->unknown_len);
            strcat (prefs->unknown, "S ");
            strcat (prefs->unknown, beg);
            strcat (prefs->unknown, "\" ");
            break;
     }

    return (unsigned char*)delim + 2;
}

/**
 * Parses one int value from the nondef string and stores
 * the value inside prefs.
 *
 * @param prefs - The pws_prefs_t object were all pref values will be stored.
 * @param buf   - The single int string nondef value.
 *
 * @returns The start of the next preference inside the nondef String.
 */
unsigned char * pws_parse_pref_int (pws_prefs_t * prefs, char * buf)
{
    char * beg = buf;

    int typ = strtol (buf, &buf, 10);
    switch (typ)
    {
        case PWS_PREFS_INT_PWLENDEFAULT:
            prefs->pwlendefault = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_INT_IDLETIMEOUT:
            prefs->idletimeout = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_INT_TREEDISPLAYSTATUSATOPEN:
            prefs->treedisplaystatusatopen = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_INT_NUMPWHISTORYDEFAULT:
            prefs->numpwhistorydefault = strtol (buf, &buf, 10);
            break;
        default:
            if (!prefs->unknown)
            {
                prefs->unknown = strdup ("");
                prefs->unknown_len = 1;
            }
            
            strtol (buf, &buf, 10); *buf = 0;
            prefs->unknown_len += strlen (beg) + 3;
            prefs->unknown = realloc (prefs->unknown, prefs->unknown_len);
            strcat (prefs->unknown, "I ");
            strcat (prefs->unknown, beg);
            strcat (prefs->unknown, " ");
            break;
    }

    return (unsigned char*)buf + 1;
}

/**
 * Parses one bool value from the nondef string and stores
 * the value inside prefs.
 *
 * @param prefs - The pws_prefs_t object were all pref values will be stored.
 * @param buf   - The single bool string nondef value.
 *
 * @returns The start of the next preference inside the nondef String.
 */
unsigned char * pws_parse_pref_bool (pws_prefs_t * prefs, char * buf)
{
    char * beg = buf;

    int typ = strtol (buf, &buf, 10);
    switch (typ)
    {
        case PWS_PREFS_BOOL_SHOWPWDEFAULT:
            prefs->showpwdefault = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_SHOWPASSWORDINTREE:
            prefs->showpasswordintree = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_SORTASCENDING:
            prefs->sortascending = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_USEDEFUSER:
            prefs->usedefuser = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_SAVEIMMEDIATELY:
            prefs->saveimmediately = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_PWUSELOWERCASE:
            prefs->pwuselowercase = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_PWUSEUPPERCASE:
            prefs->pwuseuppercase = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_PWUSEDIGITS:
            prefs->pwusedigits = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_PWUSESYSMBOLS:
            prefs->pwusesysmbols = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_PWUSEHEXDIGIT:
            prefs->pwusehexdigit = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_PWEASYVISION:
            prefs->pweasyvision = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_ISUTF8:
            prefs->isutf8 = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_DISPLAYEXPANDEDADDEDITDIALOG:
            prefs->displayexpandedaddeditdialog = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_MAINTAINDATETIMESTAMPS:
            prefs->maintaindatetimestamps = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_SAVEPASSWORDHISTORY:
            prefs->savepasswordhistory = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_SHOWNOTESDEFAULT:
            prefs->shownotesdefault = strtol (buf, &buf, 10);
            break;
        case PWS_PREFS_BOOL_SHOWUSERNAMEINTREE:
            prefs->showusernameintree = strtol (buf, &buf, 10);
            break;
        default:
            if (!prefs->unknown)
            {
                prefs->unknown = strdup ("");
                prefs->unknown_len = 1;
            }
            
            strtol (buf, &buf, 10); *buf = 0;
            prefs->unknown_len += strlen (beg) + 3;
            prefs->unknown = realloc (prefs->unknown, prefs->unknown_len);
            strcat (prefs->unknown, "B ");
            strcat (prefs->unknown, beg);
            strcat (prefs->unknown, " ");
            break;
    }

    return (unsigned char*)buf + 1;
}

/**
 * Parses the nondev value string and stores all
 * values in prefs
 *
 * @param prefs - The pws_prefs_t object were all pref values will be stored.
 * @param buf   - The nondef value string to be parsed.
 * @param len   - Length of the nondef string.
 */
void pws_parse_prefs (pws_prefs_t *prefs, unsigned char * buf, unsigned int len)
{
    unsigned char *end = buf + len;

    while (buf && ((buf+2) < end))
    {
        char typ = buf[0];
        buf += 2;
        switch (typ)
        {
            case 'B':
                buf = pws_parse_pref_bool (prefs, (char*)buf);
                break;
            case 'I':
                buf = pws_parse_pref_int (prefs, (char*)buf);
                break;
            case 'S':
                buf = pws_parse_pref_string (prefs, (char*)buf);
                break;
            default: /* invalid string, stop parsing at this point */
                return;
                break;
        }
     }
}

/**
 * Builds the password history entry data.
 *
 * @param pws_prefs_t     - The pws_prefs_t object (to get the default password history number).
 * @param pws_entry_t     - The pws_entry_t object containing the password history.
 * @param history         - The pwd history entry data will be stored inside pwd_history.
 * @param pwd_history_len - The length of pwd_history, will be stored inside pwd_history_len.
 */
void pws_pwd_history (pws_prefs_t *prefs, pws_entry_t * entry, char **pwd_history, int *pwd_history_len)
{
    *pwd_history_len = 5;
    int x; for (x = 0; x < entry->pwd_history_len; x++)
        { *pwd_history_len += entry->pwd_history[x].pwd_len + 12; }

    *pwd_history = malloc_secure (((*pwd_history_len)+1)*sizeof(char));

    sprintf (*pwd_history, "%d", entry->save_pwd_history ? 1 : 0);
    if (entry->def_pwd_max > 0)
        { sprintf (*pwd_history+1, "%02x", entry->def_pwd_max); }
    else
        { sprintf (*pwd_history+1, "%02x", prefs->numpwhistorydefault); }
    sprintf (*pwd_history+3, "%02x", entry->pwd_history_len);

    int pos = 5;
    for (x = 0; x < entry->pwd_history_len; x++)
    {
        sprintf (*pwd_history+pos, "%08x", (unsigned int)entry->pwd_history[x].timestamp);
        pos += 8;

        sprintf (*pwd_history+pos, "%04x", (unsigned int)entry->pwd_history[x].pwd_len);
        pos += 4;

        memcpy (*pwd_history+pos, entry->pwd_history[x].pwd, entry->pwd_history[x].pwd_len);
        pos += entry->pwd_history[x].pwd_len;
    }
}

/**
 * Parses the pwd_history string and stores all
 * values in entry->pwd_history.
 *
 * @param entry           - The pws_entry_t object were all history entries values will be stored.
 * @param pwd_history     - The pwd_history string to be parsed.
 * @param pwd_history_len - Length of the pwd_history string.
 */
void pws_parse_pwd_history (pws_entry_t *entry, unsigned char * pwd_history, unsigned int pwd_history_len)
{
    unsigned int max_entries;
    unsigned int num_entries;
    char buf[255];

    if (pwd_history_len > 0 && pwd_history[0] == '1') 
        { entry->save_pwd_history = 1; }
    else if (pwd_history_len > 0 && pwd_history[0] == '0') 
        { entry->save_pwd_history = 0; }
    else if (pwd_history_len > 0 && pwd_history[0] == 1) 
        { entry->save_pwd_history = 1; }
    else if (pwd_history_len > 0 && pwd_history[0] == 0) 
        { entry->save_pwd_history = 0; }
    else
        { return; }

    if (pwd_history_len > 2)
    {
        memset (buf, 0, 255);
        memcpy (buf, pwd_history+1, 2);

        max_entries = strtol(buf, NULL, 16); 
    }
    else
    {
        return;
    }

    if (pwd_history_len > 4)
    {
        memset (buf, 0, 255);
        memcpy (buf, pwd_history+3, 2);

        num_entries = strtol(buf, NULL, 16); 
    }
    else
    {
        return;
    }

    int pos = 5;
    int i; for (i = 0; i < num_entries; i++)
    {
        time_t timestamp;
        if (pwd_history_len > pos+8)
        {
            memset (buf, 0, 255);
            memcpy (buf, pwd_history+pos, 8);

            timestamp = (time_t)strtol(buf, NULL, 16); 
            pos += 8;
        }
        else
        {
            return;
        }

        unsigned len = 0;
        if (pwd_history_len > pos+4)
        {
            memset (buf, 0, 255);
            memcpy (buf, pwd_history+pos, 4);

            len = strtol(buf, NULL, 16);
            pos += 4;
        }
        else
        {
            return;
        }

        if (pwd_history_len >= pos+len)
        {
            memset (buf, 0, 255);
            memcpy (buf, pwd_history+pos, len);

            pos += len;
        }
        else
        {
            return;
        }

        entry->pwd_history = realloc (entry->pwd_history, (entry->pwd_history_len+1)*sizeof(pws_pwd_hst_entry_t));

        entry->pwd_history[entry->pwd_history_len].pwd_len = len;
        entry->pwd_history[entry->pwd_history_len].pwd     = malloc_secure (len*sizeof(char));
        entry->pwd_history[entry->pwd_history_len].timestamp = timestamp;

        memcpy (entry->pwd_history[entry->pwd_history_len].pwd, buf, len);
        entry->pwd_history_len += 1;
    }
    entry->def_pwd_max = max_entries;
}

/**
 * Appends one bool value to the nondef string.
 *
 * @param buf - The nondef string.
 * @param typ - The type of the value.
 * @param value - The value to be appended.
 *
 * @returns The new nondef String.
 */
char * pws_append_bool_pref (char * buf, unsigned int typ, unsigned int value)
{
    int siz = 7;
    if (typ > 9)  siz++;
    if (typ > 99) siz++;
    
    if (value) value = 1;

    char *b = malloc (siz*sizeof(char)); 
    sprintf (b, "B %i %i ", typ, value);

    if (!buf) { buf = malloc (siz*sizeof(char)); buf[0] = 0; }
    else      { buf = realloc (buf, (siz+strlen(buf))*sizeof(char)); }

    strcat (buf, b);

    free (b);
    return buf;
}

/**
 * Appends one int value to the nondef string.
 *
 * @param buf - The nondef string.
 * @param typ - The type of the value.
 * @param value - The value to be appended.
 *
 * @returns The new nondef String.
 */
char * pws_append_int_pref (char * buf, unsigned int typ, unsigned int value)
{
    int siz = 7;
    if (typ > 9)  siz++;
    if (typ > 99) siz++;
    
    if (value > 9)   siz++;
    if (value > 99)  siz++;
    if (value > 999) siz++;

    char *b = malloc (siz*sizeof(char)); 
    sprintf (b, "I %i %i ", typ, value);

    if (!buf) { buf = malloc (siz*sizeof(char)); buf[0] = 0; }
    else      { buf = realloc (buf, (siz+strlen(buf))*sizeof(char)); }

    strcat (buf, b);

    free (b);
    return buf;
}

/**
 * Appends one string value to the nondef string.
 *
 * @param buf - The nondef string.
 * @param typ - The type of the value.
 * @param str - The string to be appended.
 * @param len - Length of str.
 *
 * @returns The new nondef String.
 */
char * pws_append_string_pref (char * buf, unsigned int typ, char * str, unsigned int len)
{
    int siz = len + 8;
    if (typ > 9)  siz++;
    if (typ > 99) siz++;
    
    char *string = malloc ((len+1)*sizeof (char));
    strncpy (string, str, len); string[len] = '\0';

    char *b = malloc (siz*sizeof(char)); 
    sprintf (b, "S %i \"%s\" ", typ, string);

    if (!buf) { buf = malloc (siz*sizeof(char)); buf[0] = 0; }
    else      { buf = realloc (buf, (siz+strlen(buf))*sizeof(char)); }

    strcat (buf, b);

    free (string);
    free (b);
    return buf;
}

/**
 * Creates the nondef string. The string must be freed using free().
 *
 * @param prefs - The preferences to be written.
 *
 * @returns The nondef String.
 */
char * pws_nondef_string (pws_prefs_t * prefs)
{
    char * nondef = NULL;

    // boolean preferences
    if (prefs->showpwdefault != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SHOWPWDEFAULT, prefs->showpwdefault); }

    if (prefs->showpasswordintree != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SHOWPASSWORDINTREE, prefs->showpasswordintree); }

    if (prefs->sortascending != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SORTASCENDING, prefs->sortascending); }

    if (prefs->usedefuser != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_USEDEFUSER, prefs->usedefuser); }

    if (prefs->saveimmediately != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SAVEIMMEDIATELY, prefs->saveimmediately); }

    if (prefs->pwuselowercase != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_PWUSELOWERCASE, prefs->pwuselowercase); }

    if (prefs->pwuseuppercase != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_PWUSEUPPERCASE, prefs->pwuseuppercase); }

    if (prefs->pwusedigits != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_PWUSEDIGITS, prefs->pwusedigits); }

    if (prefs->pwusesysmbols != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_PWUSESYSMBOLS, prefs->pwusesysmbols); }

    if (prefs->pwusehexdigit != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_PWUSEHEXDIGIT, prefs->pwusehexdigit); }

    if (prefs->pweasyvision != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_PWEASYVISION, prefs->pweasyvision); }

    if (prefs->isutf8 != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_ISUTF8, prefs->isutf8); }

    if (prefs->displayexpandedaddeditdialog != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_DISPLAYEXPANDEDADDEDITDIALOG, prefs->displayexpandedaddeditdialog); }

    if (prefs->maintaindatetimestamps != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_MAINTAINDATETIMESTAMPS, prefs->maintaindatetimestamps); }

    if (prefs->savepasswordhistory != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SAVEPASSWORDHISTORY, prefs->savepasswordhistory); }

    if (prefs->shownotesdefault != 0)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SHOWNOTESDEFAULT, prefs->shownotesdefault); }

    if (prefs->showusernameintree != 1)
        { nondef = pws_append_bool_pref (nondef, PWS_PREFS_BOOL_SHOWUSERNAMEINTREE, prefs->showusernameintree); }

    nondef = pws_append_bool_pref (nondef, 99, 0);

    // int prefs
    if (prefs->pwlendefault != 8)
        { nondef = pws_append_int_pref (nondef, PWS_PREFS_INT_PWLENDEFAULT, prefs->pwlendefault); }

    if (prefs->idletimeout != 5)
        { nondef = pws_append_int_pref (nondef, PWS_PREFS_INT_IDLETIMEOUT, prefs->idletimeout); }

    if (prefs->treedisplaystatusatopen != 0)
        { nondef = pws_append_int_pref (nondef, PWS_PREFS_INT_TREEDISPLAYSTATUSATOPEN, prefs->treedisplaystatusatopen); }

    if (prefs->numpwhistorydefault != 3)
        { nondef = pws_append_int_pref (nondef, PWS_PREFS_INT_NUMPWHISTORYDEFAULT, prefs->numpwhistorydefault); }

    nondef = pws_append_int_pref (nondef, 99, 99);


    // string prefs
    if (prefs->defaultusername != NULL)
        { nondef = pws_append_string_pref (nondef, PWS_PREFS_STRING_DEFAULTUSERNAME, prefs->defaultusername, prefs->defaultusername_len); }

    if (prefs->defaultautootypestring != NULL)
        { nondef = pws_append_string_pref (nondef, PWS_PREFS_STRING_DEFAULTAUTOOTYPESTRING, prefs->defaultautootypestring, prefs->defaultautootypestring_len); }

    nondef = pws_append_string_pref (nondef, 99, "99", 2);

    return nondef;
}

/**
 * Sets error->error and error->str_error.
 *
 * @param pws_error_t - The pws_error_t object.
 * @param err         - The error code.
 * @param str_err     - A string describing the error code.
 *
 * @returns 0.
 */
int pws_error (pws_error_t *error, unsigned int err, const char *str_err)
{
    if (error->str_error)
        free (error->str_error);

    error->error = err;
    error->str_error = strdup (str_err);

    return 0;
}

/**
 * Returns the last pws error code.
 *
 * @param dat - The pws_context_t object that should be analyzed.
 *
 * @returns The pws error code.
 */
int pws_int_error (pws_context_t *ctx)
{
    return (ctx->error.error);
}

/**
 * Returns the string representaion of the last pws error.
 *
 * @param dat - The pws_context_t object that should be analyzed.
 *
 * @returns The pws error code.
 */
char * pws_str_error (pws_context_t *ctx)
{
    return (ctx->error.str_error);
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
