/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef PWS_H
#define PWS_H 1

#include <stdio.h>
#include <gcrypt.h>
#include "config.h"
#include "secmem.h"

#ifdef MINGW32
#include "mingw.h"
#else
#include <arpa/inet.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define PWSV2_HDR " !!!Version 2 File Format!!! Please upgrade to PasswordSafe 2.0 or later"
#define PWSV2_VER "2.0"

#define PWSV3_VER 0x0302

// pws v3 header types
#define PWS_HDR_TYP_VERSION          0x00
#define PWS_HDR_TYP_UUID             0x01
#define PWS_HDR_TYP_NONDEF           0x02
#define PWS_HDR_TYP_TREE_STATUS      0x03
#define PWS_HDR_TYP_LAST_MODIFIED    0x04
#define PWS_HDR_TYP_LAST_SAVE_AUTHOR 0x05
#define PWS_HDR_TYP_LAST_SAVE_APP    0x06
#define PWS_HDR_TYP_LAST_SAVE_UID    0x07
#define PWS_HDR_TYP_LAST_SAVE_HOST   0x08
#define PWS_HDR_TYP_DB_NAME          0x09
#define PWS_HDR_TYP_DB_DESC          0x0a
#define PWS_HDR_TYP_EOH              0xff

// pws v[2|3] entry types
#define PWS_ENTRY_TYP_UUID                   0x01
#define PWS_ENTRY_TYP_GROUP                  0x02
#define PWS_ENTRY_TYP_TITLE                  0x03
#define PWS_ENTRY_TYP_USERNAME               0x04
#define PWS_ENTRY_TYP_NOTES                  0x05
#define PWS_ENTRY_TYP_PASSWORD               0x06
#define PWS_ENTRY_TYP_CREATION_TIME          0x07
#define PWS_ENTRY_TYP_PWD_MODIFICATION_TIME  0x08
#define PWS_ENTRY_TYP_LAST_ACCESS_TIME       0x09
#define PWS_ENTRY_TYP_PASSWORD_LIFETIME      0x0a
#define PWS_ENTRY_TYP_PASSWORD_POLICY        0x0b
#define PWS_ENTRY_TYP_LAST_MODIFICATION_TIME 0x0c
#define PWS_ENTRY_TYP_URL                    0x0d
#define PWS_ENTRY_TYP_AUTOTYPE               0x0e
#define PWS_ENTRY_TYP_PASSWORD_HISTORY       0x0f
#define PWS_ENTRY_TYP_EOE                    0xff

// pws v[2|3] non default preference types
#define PWS_PREFS_BOOL_SHOWPWDEFAULT                 1
#define PWS_PREFS_BOOL_SHOWPASSWORDINTREE            2
#define PWS_PREFS_BOOL_SORTASCENDING                 3 
#define PWS_PREFS_BOOL_USEDEFUSER                    4
#define PWS_PREFS_BOOL_SAVEIMMEDIATELY               5
#define PWS_PREFS_BOOL_PWUSELOWERCASE                6
#define PWS_PREFS_BOOL_PWUSEUPPERCASE                7
#define PWS_PREFS_BOOL_PWUSEDIGITS                   8
#define PWS_PREFS_BOOL_PWUSESYSMBOLS                 9
#define PWS_PREFS_BOOL_PWUSEHEXDIGIT                10
#define PWS_PREFS_BOOL_PWEASYVISION                 11
#define PWS_PREFS_BOOL_ISUTF8                       24
#define PWS_PREFS_BOOL_DISPLAYEXPANDEDADDEDITDIALOG 27
#define PWS_PREFS_BOOL_MAINTAINDATETIMESTAMPS       28
#define PWS_PREFS_BOOL_SAVEPASSWORDHISTORY          29
#define PWS_PREFS_BOOL_SHOWNOTESDEFAULT             31
#define PWS_PREFS_BOOL_SHOWUSERNAMEINTREE           37

#define PWS_PREFS_INT_PWLENDEFAULT                  5
#define PWS_PREFS_INT_IDLETIMEOUT                   7
#define PWS_PREFS_INT_TREEDISPLAYSTATUSATOPEN      11
#define PWS_PREFS_INT_NUMPWHISTORYDEFAULT          12

#define PWS_PREFS_STRING_DEFAULTUSERNAME            3
#define PWS_PREFS_STRING_DEFAULTAUTOOTYPESTRING    10

#define PWFILE_SUCCESS      0x00
#define PWFILE_WRONG_FORMAT 0x01
#define PWFILE_WRONG_KEY    0x02
#define PWFILE_CORRUPT      0x03
#define PWFILE_WRITE_ERROR  0x04
#define PWFILE_GCRYPT_ERROR 0x05

typedef struct PWS_PREFS
{
    int  showpwdefault;
    int  showpasswordintree;
    int  sortascending;
    int  usedefuser;
    int  saveimmediately;
    int  pwuselowercase;
    int  pwuseuppercase;
    int  pwusedigits;
    int  pwusesysmbols;
    int  pwusehexdigit;
    int  pweasyvision;
    int  isutf8;
    int  displayexpandedaddeditdialog;
    int  maintaindatetimestamps;
    int  savepasswordhistory;
    int  shownotesdefault;
    int  showusernameintree;

    int pwlendefault;
    int idletimeout;
    int treedisplaystatusatopen;
    int numpwhistorydefault;

    char * defaultusername;
    int    defaultusername_len;
    char * defaultautootypestring;
    int    defaultautootypestring_len;

    char * unknown;
    int    unknown_len;
} pws_prefs_t;

typedef struct PWS3_HDR
{
    char  *version;
    int    version_len;
    unsigned char  *uuid;
    int    uuid_len;
    char  *tree_status;
    int    tree_status_len;
    char  *last_app;
    int    last_app_len;
    char  *last_uid;
    int    last_uid_len;
    char  *last_host;
    int    last_host_len;
    char  *db_name;
    int    db_name_len;
    char  *db_desc;
    int    db_desc_len;
    char  *unknown;
    int    unknown_len;
    time_t last_modified;
} pws3_hdr_t;

typedef struct PWS2_HEADER
{
    char  *name;
    int    name_len;
    char  *password;
    int    password_len;
    char  *notes;
    int    notes_len;
} pws2_hdr_t;


typedef struct PWS_PWD_HST_ENTRY
{
    time_t timestamp;
    char * pwd;
    int pwd_len;
} pws_pwd_hst_entry_t;

typedef struct PWS_ENTRY
{
    unsigned char  *uuid;
    int    uuid_len;
    char  *group;
    int    group_len;
    char  *title;
    int    title_len;
    char  *username;
    int    username_len;
    char  *notes;
    int    notes_len;
    char  *pwd;
    int    pwd_len;
    char  *pwd_policy;
    int    pwd_policy_len;
    char  *url;
    int    url_len;
    char  *autotype;
    int    autotype_len;
    char  *unknown;
    int    unknown_len;
    time_t ctime;
    time_t pwd_mtime;
    time_t mtime;
    time_t atime;
    time_t ltime;
    pws_pwd_hst_entry_t *pwd_history;
    int pwd_history_len;
    int save_pwd_history;
    int def_pwd_max;
} pws_entry_t;

typedef struct PWS_ERROR
{
    int  error;
    char *str_error;
} pws_error_t;

typedef struct PWS_DATA
{
    pws2_hdr_t     hdr2;
    pws3_hdr_t     hdr3;
    pws_prefs_t    prefs;
    pws_entry_t **entries;
    int            num_entries;
} pws_data_t;

typedef struct PWS3
{
    unsigned int   iter;
    char          *p;
    unsigned char *ptag;
    unsigned char *hptag;
    unsigned char *salt;
    unsigned char *K;
    unsigned char *L;
    unsigned char *IV;
    unsigned char *HMAC;
    unsigned char *data;
    unsigned int   data_len;
    gcry_cipher_hd_t *cbc_ctx;
} pws3_t;

typedef struct PWS2
{
    char          *p;
    unsigned char *key;
    unsigned char *rnd;
    unsigned char *hrnd;
    unsigned char *salt;
    unsigned char *IP;
} pws2_t;

typedef struct PWS_BUFFER
{
    unsigned char *data;
    unsigned int   data_len;
} pws_buffer_t;

typedef struct PWS_CONTEXT
{
    pws2_t pws2;
    pws3_t pws3;
    pws_error_t error;
} pws_context_t;

/**
 * Some descriptive error messages.
 */
#ifdef MINGW32
EXPORT
#endif
extern const char *STR_ERROR[];

/**
 * Frees/trashes all memory used by dat.
 *
 * @param dat - The pws_data_t object to be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_release_data (pws_data_t *dat);

/**
 * Sets all members of dat to NULL.
 *
 * @param dat - The pws_data_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_init_data (pws_data_t *dat);

/**
 * Sets all members of prefs to default values.
 *
 * @param prefs - The pws_prefs_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_init_prefs (pws_prefs_t *prefs);

/**
 * Frees/trashes all memory used by prefs.
 *
 * @param prefs - The pws_prefs_t object to be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_release_prefs (pws_prefs_t *prefs);

/**
 * Sets all members of entry to NULL.
 *
 * @param entry - The pws_entry_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_init_entry (pws_entry_t *entry);

/**
 * Frees/trashes all memory used by entry.
 *
 * @param entry - The pws_entry_t object to be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_release_entry (pws_entry_t *entry);

/**
 * Sets all members of hdr to NULL.
 *
 * @param hdr - The pws2_hdr_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws2_init_hdr (pws2_hdr_t *hdr);

/**
 * Frees/trashes all memory used by hdr.
 *
 * @param hdr - The pws2_hdr_t object to be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws2_release_hdr (pws2_hdr_t *hdr);

/**
 * Sets all members of hdr to NULL.
 *
 * @param hdr - The pws3_hdr_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws3_init_hdr (pws3_hdr_t *hdr);

/**
 * Frees/trashes all memory used by hdr.
 *
 * @param hdr - The pws3_hdr_t object to be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws3_release_hdr (pws3_hdr_t *hdr);

/**
 * Parses the nondev value string and stores all
 * values in prefs
 *
 * @param prefs - The pws_prefs_t object were all pref values will be stored.
 * @param buf   - The nondef value string to be parsed.
 * @param len   - Length of the nondef string.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_parse_prefs (pws_prefs_t *prefs, unsigned char * buf, unsigned int len);

/**
 * Creates the nondef string. The string must be freed using free().
 *
 * @param prefs - The preferences to be written.
 *
 * @returns The nondef String.
 */
#ifdef MINGW32
EXPORT
#endif
char * pws_nondef_string (pws_prefs_t * prefs);

/**
 * Parses the pwd_history string and stores all
 * values in entry->pwd_history.
 *
 * @param entry           - The pws_entry_t object were all history entries values will be stored.
 * @param pwd_history     - The pwd_history string to be parsed.
 * @param pwd_history_len - Length of the pwd_history string.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_parse_pwd_history (pws_entry_t *entry, unsigned char * pwd_history, unsigned int pwd_history_len);

/**
 * Builds the password history entry data.
 *
 * @param pws_prefs_t - The pws_prefs_t object (to get the default password history number).
 * @param pws_entry_t - The pws_entry_t object containing the password history.
 * @param history     - The pwd history entry data will be stored inside pwd_history.
 * @param pwd_history_len - The length of pwd_history, will be stored inside pwd_history_len.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_pwd_history (pws_prefs_t *prefs, pws_entry_t * entry, char **pwd_history, int *pwd_history_len);

/**
 * Sets error->error and error->str_error.
 *
 * @param pws_error_t - The pws_error_t object.
 * @param err         - The error code.
 * @param str_err     - A string describing the error code.
 *
 * @returns 0.
 */
#ifdef MINGW32
EXPORT
#endif
int pws_error (pws_error_t *error, unsigned int err, const char *str_err);

/**
 * Initializes the pws_context_t object,
 * Must be called before using any other decrypting functions.
 *
 * @param ctx - The pws_context_t context object containing the context that should be initialized.
 * @param p   - The passphrase.
 * @param enc_only - If set to 1, only the pwd hash will be hold in memory, the pwd itself
 *                   will be cleaned immediately. Only valid if this context will be used
 *                   only for encrypting purposes. If this context is used for decryption also,
 *                   dec_only must be set to 0.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_init_context (pws_context_t *ctx, char *p, int enc_only);

/**
 * Frees/trashes all ressources used by the pws_context_t object.
 *
 * @param ctx - The pws_context_t context object containing the context that should be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws_release_context (pws_context_t *ctx);

/**
 * Forward declaration (defined in pws3crypt.h)
 *
 * Initializes the pws3_t context,
 * Must be called before using any other
 * decrypting functions.
 *
 * @param dat - The pws3_t context object containing the context that should be initialized.
 * @param p   - The passphrase.
 * @param dec_only - If set to 1, only the pwd hash will be hold in memory, the pwd itself
 *                   will be cleaned immediately. Only valid if this context will be used
 *                   only for encrypting purposes. If this context is used for decryption also,
 *                   dec_only must be set to 0.
 * @param error - The pws_error_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws3_init_ctx (pws3_t *ctx, char *p, int enc_only, pws_error_t *error);

/**
 * Forward declaration (defined in pws3crypt.h)
 *
 * Frees/trashes all ressources used by the pws3_t context.
 *
 * @param ctx - The pws3_t context object containing the context that should be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws3_release_ctx (pws3_t *ctx);

/**
 * Forward declaration (defined in pws2crypt.h)
 *
 * Initializes the pws2_t context.
 * Must be called before using any other
 * decrypting functions.
 *
 * @param ctx - The pws2_t context object that should be initialized.
 * @param p   - The passphrase.
 * @param enc_only - If set to 1, only the pwd hash will be hold in memory, the pwd itself
 *                   will be cleaned immediately. Only valid if this context will be used
 *                   only for encrypting purposes. If this context is used for decryption also,
 *                   dec_only must be set to 0.
 * @param error - The pws_error_t object.
 */
#ifdef MINGW32
EXPORT
#endif
void pws2_init_ctx (pws2_t *ctx, char *p, int enc_only, pws_error_t *error);

/**
 * Forward declaration (defined in pws2crypt.h)
 *
 * Frees/trashes all ressources used by the pws2_t context.
 *
 * @param ctx - The pws2_t context object that should be freed.
 */
#ifdef MINGW32
EXPORT
#endif
void pws2_release_ctx (pws2_t *ctx);

/**
 * Returns the last pws error code.
 *
 * @param dat - The pws_context_t object that should be analyzed.
 *
 * @returns The pws error code.
 */
#ifdef MINGW32
EXPORT
#endif
int pws_int_error (pws_context_t *ctx);

/**
 * Returns the string representaion of the last pws error.
 *
 * @param dat - The pws_context_t object that should be analyzed.
 *
 * @returns The pws error code.
 */
#ifdef MINGW32
EXPORT
#endif
char * pws_str_error (pws_context_t *ctx);

#ifdef __cplusplus
}
#endif

#endif

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
