/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pws2.h"

/**
 * Reads the pws v2 header and stores all recognized typed entries inside hdr.
 * ptr must point to the beginning of the unencrypted data block
 * (pws_buffer_t->data).
 *
 * ptr is positioned at the beginning of the next Password-Entry-Block
 * and returned.
 *
 * @param dat - The pws_data_t object to be filled.
 * @param ptr - Points to the start of the (decrypted) header block.
 * @param end - Points to the last valid byte inside ptr.
 *
 * @returns The start off the next typed field.
 */
unsigned char * pws2_read_header (pws_data_t *dat, unsigned char *ptr, unsigned char *end)
{
    unsigned int len;
    // username: "!!!Version 2 File Format!!!"
    if ((ptr+4) > end) { return (NULL); }
    memcpy (&len, ptr, sizeof (int));
    if ((ptr+PWS2_BLOCK_SIZE+len) > end) { return (NULL); }

    ptr += PWS2_BLOCK_SIZE;
    dat->hdr2.name = malloc (len*sizeof(char));
    dat->hdr2.name_len = len;
    memcpy (dat->hdr2.name, ptr, len);

    ptr += len;
    if (len % PWS2_BLOCK_SIZE)
        { ptr += (PWS2_BLOCK_SIZE-(len % PWS2_BLOCK_SIZE)); }

    // password: pws v2 version string
    if ((ptr+4) > end) { return (NULL); }
    memcpy (&len, ptr, sizeof (int));
    if ((ptr+PWS2_BLOCK_SIZE+len) > end) { return (NULL); }

    ptr += PWS2_BLOCK_SIZE;
    dat->hdr2.password = malloc (len*sizeof(char));
    dat->hdr2.password_len = len;
    memcpy (dat->hdr2.password, ptr, len);

    ptr += len;
    if (len % PWS2_BLOCK_SIZE)
        { ptr += (PWS2_BLOCK_SIZE-(len % PWS2_BLOCK_SIZE)); }

    // notes
    if ((ptr+4) > end) { return (NULL); }
    memcpy (&len, ptr, sizeof (int));
    if ((ptr+PWS2_BLOCK_SIZE+len) > end) { return (NULL); }

    ptr += PWS2_BLOCK_SIZE;
    dat->hdr2.notes = malloc (len*sizeof(char));
    dat->hdr2.notes_len = len;
    memcpy (dat->hdr2.notes, ptr, len);

    ptr += len;
    if (len % PWS2_BLOCK_SIZE)
        { ptr += (PWS2_BLOCK_SIZE-(len % PWS2_BLOCK_SIZE)); }

    if (!len)
        { ptr += PWS2_BLOCK_SIZE; }

    pws_parse_prefs (&dat->prefs, (unsigned char*)dat->hdr2.notes, len+1);

    return (ptr);
}

/**
 * Reads one PWS2 entry block and stores all recognized entries inside entry.
 * ptr should point to the beginning of the next unencrypted entry block.
 *
 * ptr is positioned at the beginning of the next Password-Entry-Block
 * and returned.
 *
 * @param entry - The pws_entry_t object to be filled.
 * @param ptr   - Points to the start of the (decrypted) entry block.
 * @param end   - Points to the last valid byte inside ptr.
 *
 * @returns The start off the next typed field.
 */
unsigned char * pws2_read_entry (pws_entry_t *entry, unsigned char *ptr, unsigned char *end)
{
    unsigned int len, l;
    unsigned short w1, w2;

    if ((ptr+5) > end) { return (NULL); }
    memcpy (&len, ptr, sizeof (int));
    unsigned int typ = ptr[4];
    if ((ptr+PWS2_BLOCK_SIZE+len) > end) { return (NULL); }

    ptr += PWS2_BLOCK_SIZE;
    while (typ != PWS_ENTRY_TYP_EOE)
    {
        switch (typ)
        {
            case PWS_ENTRY_TYP_UUID: 
                entry->uuid = malloc_secure (len*sizeof(char));
                entry->uuid_len = len;
                memcpy (entry->uuid, ptr, len);
                if (len > (sizeof(int)+(2*sizeof(short))))
                {
                    memcpy (&l,  entry->uuid,   sizeof(int));
                    memcpy (&w1, entry->uuid+4, sizeof(short));
                    memcpy (&w2, entry->uuid+6, sizeof(short));
                    l = ntohl (l); w1 = ntohs (w1); w2 = ntohs (w2);
                    memcpy (entry->uuid,    &l, sizeof(int));
                    memcpy (entry->uuid+4, &w1, sizeof(short));
                    memcpy (entry->uuid+6, &w2, sizeof(short));
                }
                break;
            case PWS_ENTRY_TYP_GROUP: 
                entry->group = malloc_secure (len*sizeof(char));
                entry->group_len = len;
                memcpy (entry->group, ptr, len);
                break;
            case PWS_ENTRY_TYP_TITLE: 
                entry->title = malloc_secure (len*sizeof(char));
                entry->title_len = len;
                memcpy (entry->title, ptr, len);
                break;
            case PWS_ENTRY_TYP_USERNAME: 
                entry->username = malloc_secure (len*sizeof(char));
                entry->username_len = len;
                memcpy (entry->username, ptr, len);
                break;
            case PWS_ENTRY_TYP_NOTES: 
                entry->notes = malloc_secure (len*sizeof(char));
                entry->notes_len = len;
                memcpy (entry->notes, ptr, len);
                break;
            case PWS_ENTRY_TYP_PASSWORD: 
                entry->pwd = malloc_secure (len*sizeof(char));
                entry->pwd_len = len;
                memcpy (entry->pwd, ptr, len);
                break;
            case PWS_ENTRY_TYP_CREATION_TIME: 
                memcpy ((void*)&entry->ctime, ptr, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_PWD_MODIFICATION_TIME: 
                memcpy ((void*)&entry->pwd_mtime, ptr, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_LAST_ACCESS_TIME: 
                memcpy ((void*)&entry->atime, ptr, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_PASSWORD_LIFETIME: 
                memcpy ((void*)&entry->ltime, ptr, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_PASSWORD_POLICY: 
                entry->pwd_policy = malloc_secure (len*sizeof(char));
                entry->pwd_policy_len = len;
                memcpy (entry->pwd_policy, ptr, len);
                break;
            case PWS_ENTRY_TYP_LAST_MODIFICATION_TIME: 
                memcpy ((void*)&entry->mtime, ptr, sizeof(time_t));
                break;
        }

        ptr += len;
        if (len % PWS2_BLOCK_SIZE)
            { ptr += (PWS2_BLOCK_SIZE-(len % PWS2_BLOCK_SIZE)); }

        if (!len)
            { ptr += PWS2_BLOCK_SIZE; }

        if ((ptr+5) > end) { return (NULL); }
        memcpy (&len, ptr, sizeof (int));
        typ = ptr[4];
        if ((ptr+PWS2_BLOCK_SIZE+len) > end) { return (NULL); }

        ptr += PWS2_BLOCK_SIZE;
    }

    return (ptr+PWS2_BLOCK_SIZE);
}

/**
 * Decrypts the pws v2 file f. The header data and all entries are 
 * stored inside dat.
 *
 * pws_init_data() must be called before using this function.
 *
 * @param dat - The pws_data_t object to be filled.
 * @param ctx - The pws_context_t context object.
 * @param f   - The pws v2 FILE object.
 *
 * @returns 1 on success, otherwise 1 and error->error will be set to the corresponding error code.
 */
int pws2_decrypt (pws_data_t *dat, pws_context_t *ctx, FILE *f)
{
    pws_buffer_t buf;
    buf.data     = NULL;
    buf.data_len = 0;

    if (!pws2_dec_file (&ctx->pws2, &buf, f, &ctx->error))
    {
        free_secure (buf.data);
        return 0;
    }

    unsigned char * ptr = pws2_read_header (dat, buf.data, buf.data+buf.data_len);
    if (!ptr) 
    { 
        ctx->error.error = PWFILE_CORRUPT;
        ctx->error.str_error = strdup(STR_ERROR[PWFILE_CORRUPT]);

        free_secure (buf.data);
        return 0;
    }

    while (ptr != (buf.data+buf.data_len))
    {
        dat->num_entries++;
        dat->entries = realloc (dat->entries, dat->num_entries*sizeof(pws_entry_t*));
        dat->entries[dat->num_entries-1] = malloc (sizeof(pws_entry_t));
        pws_init_entry (dat->entries[dat->num_entries-1]);

        ptr = pws2_read_entry (dat->entries[dat->num_entries-1], ptr, buf.data+buf.data_len);
        if (!ptr) 
        { 
            ctx->error.error = PWFILE_CORRUPT;
            ctx->error.str_error = strdup(STR_ERROR[PWFILE_CORRUPT]);

            free_secure (buf.data);
            return 0;
        }
    }

    free_secure (buf.data);
    return 1;
}

/**
 * Writes one typed field to the buffer ptr.
 *
 * @param ptr      - The buffer to hold the typed entry.
 * @param data     - The data to be written.
 * @param data_len - The size of data.
 * @param typ      - The type of the typed entry.
 */
void pws2_write_typed_field (unsigned char *ptr, void *data, unsigned int data_len, unsigned char typ)
{
    memcpy (ptr, (void*)&data_len, sizeof(int));
    memcpy (ptr+4, &typ, 1);
    memcpy (ptr+PWS2_BLOCK_SIZE, data, data_len);
}

/**
 * Calculate how many bytes will be needed to store dat inside a pws_buffer_t object.
 *
 * @param dat - The password entries.
 *
 * @returns The size in bytes that will be consumed by dat if stored inside pws_buffer_t.
 */
unsigned long pws2_calc_buf_size (pws_data_t *dat)
{
    unsigned long siz = 0;
    
    // header bytes
    if (strlen (PWSV2_HDR) % PWS2_BLOCK_SIZE)
        { siz += (PWS2_BLOCK_SIZE-(strlen (PWSV2_HDR) % PWS2_BLOCK_SIZE)); }
    siz += strlen (PWSV2_HDR) + PWS2_BLOCK_SIZE;

    if (strlen (PWSV2_VER) % PWS2_BLOCK_SIZE)
        { siz += (PWS2_BLOCK_SIZE-(strlen (PWSV2_VER)  % PWS2_BLOCK_SIZE)); }
    siz += strlen (PWSV2_VER) + PWS2_BLOCK_SIZE;

    char * nondef = pws_nondef_string (&dat->prefs);
    if (nondef)
    {
        if (strlen (nondef) % PWS2_BLOCK_SIZE)
            { siz += (PWS2_BLOCK_SIZE-(strlen (nondef) % PWS2_BLOCK_SIZE)); }
        else if (!strlen (nondef))
            { siz += PWS2_BLOCK_SIZE; }

        siz += strlen (nondef) + PWS2_BLOCK_SIZE; 
        free (nondef);
    }

    // entry bytes
    int x; for (x = 0; x < dat->num_entries; x++)
    {
        // typed field: uuid
        if (dat->entries[x]->uuid)
        {
            siz += dat->entries[x]->uuid_len + PWS2_BLOCK_SIZE;
            if (dat->entries[x]->uuid_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->uuid_len%PWS2_BLOCK_SIZE);
        }

        // typed field: group
        if (dat->entries[x]->group)
        {
            siz += dat->entries[x]->group_len+PWS2_BLOCK_SIZE;
            if (dat->entries[x]->group_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->group_len%PWS2_BLOCK_SIZE);
        }

        // typed field: title
        if (dat->entries[x]->title)
        {
            siz += dat->entries[x]->title_len+PWS2_BLOCK_SIZE;
            if (dat->entries[x]->title_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->title_len%PWS2_BLOCK_SIZE);
        }

        // typed field: username
        if (dat->entries[x]->username)
        {
            siz += dat->entries[x]->username_len+PWS2_BLOCK_SIZE;
            if (dat->entries[x]->username_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->username_len%PWS2_BLOCK_SIZE);
        }

        // typed field: notes
        if (dat->entries[x]->notes)
        {
            siz += dat->entries[x]->notes_len+PWS2_BLOCK_SIZE;
            if (dat->entries[x]->notes_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->notes_len%PWS2_BLOCK_SIZE);
        }

        // typed field: pwd
        if (dat->entries[x]->pwd)
        {
            siz += dat->entries[x]->pwd_len+PWS2_BLOCK_SIZE;
            if (dat->entries[x]->pwd_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->pwd_len%PWS2_BLOCK_SIZE);
        }

        // typed field: policy
        if (dat->entries[x]->pwd_policy)
        {
            siz += dat->entries[x]->pwd_policy_len+PWS2_BLOCK_SIZE;
            if (dat->entries[x]->pwd_policy_len%PWS2_BLOCK_SIZE)
                siz += PWS2_BLOCK_SIZE-(dat->entries[x]->pwd_policy_len%PWS2_BLOCK_SIZE);
        }

        // typed field: ctime
        if (dat->entries[x]->ctime)
        {
            siz += sizeof(time_t)+PWS2_BLOCK_SIZE;
            siz += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
        }

        // typed field: pwd_mtime
        if (dat->entries[x]->pwd_mtime)
        {
            siz += sizeof(time_t)+PWS2_BLOCK_SIZE;
            siz += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
        }

        // typed field: mtime
        if (dat->entries[x]->mtime)
        {
            siz += sizeof(time_t)+PWS2_BLOCK_SIZE;
            siz += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
        }

        // typed field: atime
        if (dat->entries[x]->atime)
        {
            siz += sizeof(time_t)+PWS2_BLOCK_SIZE;
            siz += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
        }

        // typed field: ltime
        if (dat->entries[x]->ltime)
        {
            siz += sizeof(time_t)+PWS2_BLOCK_SIZE;
            siz += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
        }

        // typed field: End of PWS_ENTRY
        siz += 2*PWS2_BLOCK_SIZE;
    }
    
    return siz;
}

/**
 * Writes the Header Block to buf->data.
 *
 * @param dat - The pws_data_t object.
 * @param buf - The pws_buffer_t object.
 *
 * @returns The new position in buf.
 */
unsigned long pws2_write_header (pws_data_t *dat, pws_buffer_t *buf)
{
    // maybe this header was already filled by pws2_read_header
    free (dat->hdr2.name);
    free (dat->hdr2.password);
    dat->hdr2.name         = malloc (strlen (PWSV2_HDR));
    dat->hdr2.name_len     = strlen (PWSV2_HDR);
    dat->hdr2.password     = malloc (strlen (PWSV2_HDR));
    dat->hdr2.password_len = strlen (PWSV2_VER);

    // PWSV2 Header Data
    memcpy (dat->hdr2.name, PWSV2_HDR, strlen (PWSV2_HDR));
    memcpy (dat->hdr2.password, PWSV2_VER, strlen (PWSV2_VER));

    // username: "!!!Version 2 File Format!!!"
    pws2_write_typed_field (buf->data, dat->hdr2.name, dat->hdr2.name_len, 0);

    unsigned long pos = dat->hdr2.name_len + PWS2_BLOCK_SIZE;
    if (dat->hdr2.name_len % PWS2_BLOCK_SIZE)
        { pos += (PWS2_BLOCK_SIZE-(dat->hdr2.name_len % PWS2_BLOCK_SIZE)); }


    // password: pws v2 version string
    pws2_write_typed_field (buf->data+pos, dat->hdr2.password, dat->hdr2.password_len, 0);

    pos += dat->hdr2.password_len + PWS2_BLOCK_SIZE;
    if (dat->hdr2.password_len % PWS2_BLOCK_SIZE)
        { pos += (PWS2_BLOCK_SIZE-(dat->hdr2.password_len % PWS2_BLOCK_SIZE)); }

    // notes
    char * nondef = pws_nondef_string (&dat->prefs);
    if (nondef)
    { 
        free (dat->hdr2.notes);
        dat->hdr2.notes = nondef;
        dat->hdr2.notes_len = strlen (nondef);
    }

    pws2_write_typed_field (buf->data+pos, dat->hdr2.notes, dat->hdr2.notes_len, 0);

    pos += dat->hdr2.notes_len + PWS2_BLOCK_SIZE;
    if (dat->hdr2.notes_len % PWS2_BLOCK_SIZE)
        { pos += (PWS2_BLOCK_SIZE-(dat->hdr2.notes_len % PWS2_BLOCK_SIZE)); }
    if (!dat->hdr2.notes_len)
        { pos += PWS2_BLOCK_SIZE; }

    return pos;
}

/**
 * Writes one entry Block to the end of buf->data.
 *
 * @param buf   - The pws_buffer_t object.
 * @param entry - The entry that should be written.
 */
unsigned long pws2_write_entry (pws_buffer_t *buf, pws_entry_t *entry, unsigned long pos)
{
    // typed field: uuid
    if (entry->uuid)
    {
        pws2_write_typed_field (buf->data+pos, entry->uuid, entry->uuid_len, PWS_ENTRY_TYP_UUID);

        pos += entry->uuid_len + PWS2_BLOCK_SIZE;
        if (entry->uuid_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->uuid_len%PWS2_BLOCK_SIZE);
    }

    // typed field: group
    if (entry->group)
    {
        pws2_write_typed_field (buf->data+pos, entry->group, entry->group_len, PWS_ENTRY_TYP_GROUP);

        pos += entry->group_len+PWS2_BLOCK_SIZE;
        if (entry->group_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->group_len%PWS2_BLOCK_SIZE);
    }

    // typed field: title
    if (entry->title)
    {
        pws2_write_typed_field (buf->data+pos, entry->title, entry->title_len, PWS_ENTRY_TYP_TITLE);

        pos += entry->title_len+PWS2_BLOCK_SIZE;
        if (entry->title_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->title_len%PWS2_BLOCK_SIZE);
    }

    // typed field: username
    if (entry->username)
    {
        pws2_write_typed_field (buf->data+pos, entry->username, entry->username_len, PWS_ENTRY_TYP_USERNAME);

        pos += entry->username_len+PWS2_BLOCK_SIZE;
        if (entry->username_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->username_len%PWS2_BLOCK_SIZE);
    }

    // typed field: notes
    if (entry->notes)
    {
        pws2_write_typed_field (buf->data+pos, entry->notes, entry->notes_len, PWS_ENTRY_TYP_NOTES);

        pos += entry->notes_len+PWS2_BLOCK_SIZE;
        if (entry->notes_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->notes_len%PWS2_BLOCK_SIZE);
    }

    // typed field: pwd
    if (entry->pwd)
    {
        pws2_write_typed_field (buf->data+pos, entry->pwd, entry->pwd_len, PWS_ENTRY_TYP_PASSWORD);

        pos += entry->pwd_len+PWS2_BLOCK_SIZE;
        if (entry->pwd_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->pwd_len%PWS2_BLOCK_SIZE);
    }

    // typed field: policy
    if (entry->pwd_policy)
    {
        pws2_write_typed_field (buf->data+pos, entry->pwd_policy, entry->pwd_policy_len, PWS_ENTRY_TYP_PASSWORD_POLICY);

        pos += entry->pwd_policy_len+PWS2_BLOCK_SIZE;
        if (entry->pwd_policy_len%PWS2_BLOCK_SIZE)
            pos += PWS2_BLOCK_SIZE-(entry->pwd_policy_len%PWS2_BLOCK_SIZE);
    }

    // typed field: ctime
    if (entry->ctime)
    {
        pws2_write_typed_field (buf->data+pos, &entry->ctime, sizeof(time_t), PWS_ENTRY_TYP_CREATION_TIME);

        pos += sizeof(time_t)+PWS2_BLOCK_SIZE;
        pos += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
    }

    // typed field: pwd_mtime
    if (entry->pwd_mtime)
    {
        pws2_write_typed_field (buf->data+pos, &entry->pwd_mtime, sizeof(time_t), PWS_ENTRY_TYP_PWD_MODIFICATION_TIME);

        pos += sizeof(time_t)+PWS2_BLOCK_SIZE;
        pos += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
    }

    // typed field: mtime
    if (entry->mtime)
    {
        pws2_write_typed_field (buf->data+pos, &entry->mtime, sizeof(time_t), PWS_ENTRY_TYP_LAST_MODIFICATION_TIME);

        pos += sizeof(time_t)+PWS2_BLOCK_SIZE;
        pos += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
    }

    // typed field: atime
    if (entry->atime)
    {
        pws2_write_typed_field (buf->data+pos, &entry->atime, sizeof(time_t), PWS_ENTRY_TYP_LAST_ACCESS_TIME);

        pos += sizeof(time_t)+PWS2_BLOCK_SIZE;
        pos += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
    }

    // typed field: ltime
    if (entry->ltime)
    {
        pws2_write_typed_field (buf->data+pos, &entry->ltime, sizeof(time_t), PWS_ENTRY_TYP_PASSWORD_LIFETIME);

        pos += sizeof(time_t)+PWS2_BLOCK_SIZE;
        pos += PWS2_BLOCK_SIZE-(sizeof(time_t)%PWS2_BLOCK_SIZE);
    }

    // typed field: End of PWS_ENTRY
    pws2_write_typed_field (buf->data+pos, NULL, 0, PWS_ENTRY_TYP_EOE);
    pos += 2*PWS2_BLOCK_SIZE;

    return pos;
}

/**
 * Fill buf->data with all entries and the header inside dat.
 *
 * @param buf - The pws_buffer_t object.
 * @param dat - The header and the entries that should be written to buf->data.
 */
void pws2_write_data_buffer (pws_buffer_t *buf, pws_data_t *dat)
{
    buf->data_len = pws2_calc_buf_size (dat);
    buf->data = malloc_secure (buf->data_len*sizeof(char));

    unsigned long pos = pws2_write_header (dat, buf);
    int x; for (x = 0; x < dat->num_entries; x++)
    {
        pos = pws2_write_entry (buf, dat->entries[x], pos);
    }
}

/**
 * Encrypt all data inside dat to the file f.
 *
 * @param dat - The pws_data_t object, holding all passowrds/notes/header information etc.
 * @param ctx - The pws_context_t context object.
 * @param f   - The FILE object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws2_encrypt (pws_data_t *dat, pws_context_t *ctx, FILE *f)
{
    pws_buffer_t buf;
    buf.data     = NULL;
    buf.data_len = 0;

    pws2_write_data_buffer (&buf, dat);

    if (!pws2_enc_file (&ctx->pws2, &buf, f, &ctx->error)) 
    {
        free_secure (buf.data);
        return 0;
    }

    free_secure (buf.data);
    return 1;
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
