/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pws2crypt.h"
#include "init.h"

/**
 * Clears the passphrase.
 *
 * @param ctx - The initialized pws2_t object.
 */
void pws2_clear_passphrase (pws2_t *ctx)
{
    free_secure (ctx->p);
    ctx->p = NULL;
}

/**
 * Initializes the pws2_t context.
 * Must be called before using any other
 * decrypting functions.
 *
 * @param ctx - The pws2_t context object that should be initialized.
 * @param p   - The passphrase.
 * @param enc_only - If set to 1, only the pwd hash will be hold in memory, the pwd itself
 *                   will be cleaned immediately. Only valid if this context will be used
 *                   only for encrypting purposes. If this context is used for decryption also,
 *                   dec_only must be set to 0.
 * @param error - The pws_error_t object.
 */
void pws2_init_ctx (pws2_t *ctx, char *p, int enc_only, pws_error_t *error)
{
    ctx->rnd       = malloc (PWS2_BLOCK_SIZE*sizeof(char));
    ctx->hrnd      = malloc (PWS2_MAC_SIZE*sizeof(char));
    ctx->salt      = malloc (PWS2_MAC_SIZE*sizeof(char));
    ctx->IP        = malloc (PWS2_IP_SIZE *sizeof(char));
    ctx->key       = malloc_secure (PWS2_MAC_SIZE*sizeof(char));

    ctx->p         = NULL;

    gcry_create_nonce (ctx->rnd,  PWS2_BLOCK_SIZE*sizeof(char));
    gcry_create_nonce (ctx->salt, PWS2_MAC_SIZE*sizeof(char));
    gcry_create_nonce (ctx->IP,   PWS2_IP_SIZE*sizeof(char));
    
    // clone p
    ctx->p = strdup_secure (p);

    // calculate H(RND)
    pws2_calc_hrnd (ctx, ctx->hrnd, error);

    // calculate the key used for encryption
    pws2_calc_key (ctx);

    if (enc_only)
    {
        pws2_clear_passphrase (ctx);
    }
}

/**
 * Frees/trashes all ressources used by the pws2_t context.
 *
 * @param ctx - The pws2_t context object that should be freed.
 */
void pws2_release_ctx (pws2_t *ctx)
{
    free (ctx->rnd);
    free (ctx->hrnd);
    free (ctx->salt);
    free (ctx->IP);

    free_secure (ctx->p);
    free_secure (ctx->key);
}

/**
 * Used for PWSv2 "CBC" encryption mode. Does a xor operation
 * on the two buffers: buf1 ^= buf2.
 *
 * @param buf1 - Buffer 1.
 * @param buf2 - Buffer 2.
 * @param len  - Then len of buf1.
 */
void pws2_xormem (unsigned char * buf1, unsigned char * buf2, size_t len)
{
    unsigned int x; 
    for (x = 0; x < len; x++)
        { buf1[x] ^= buf2[x]; }
}

/**
 * Swaps all bytes in buf.
 *
 * Before: 1 2 3 4 5 6 7 8
 * After:  4 3 2 1 8 7 6 5
 *
 * @param buf - The buffer to be swapped.
 * @param len - The size of buffer (must be even).
 */
void pws2_swap_bytes (unsigned char * buf, size_t len)
{
    unsigned char * tmp_buf = malloc (len);

    unsigned int x; 
    for (x = 0; x < (len/2); x++)
    {
        tmp_buf[x]         = buf[(len/2)-1-x];
        tmp_buf[x+(len/2)] = buf[(len-1)-x];
    }

    memcpy (buf, tmp_buf, len);
    free (tmp_buf);
}

/**
 * Calculates H(RND) and stores the result in hrnd. pws2->rnd and
 * pws2->p needs to be already set.
 *
 * @param pws2  - The pws2_t object.
 * @param hrnd  - The buffer where H(RND) will be stored.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0.
 */
int pws2_calc_hrnd (pws2_t *pws2, unsigned char * hrnd, pws_error_t * error)
{
    unsigned char * buf = malloc_secure ((strlen(pws2->p)+PWS2_BLOCK_SIZE+2)*sizeof(char));
    unsigned char * tmp_salt = malloc_secure (PWS2_MAC_SIZE*sizeof(char));
    unsigned char * tmp_rnd = malloc_secure (PWS2_MAC_SIZE*sizeof(char));

    memcpy (tmp_rnd, pws2->rnd, PWS2_BLOCK_SIZE);
    pws2_swap_bytes (tmp_rnd, PWS2_BLOCK_SIZE);
    memset (tmp_rnd+PWS2_BLOCK_SIZE, 0x00, 2);

    // tempSalt
    memcpy (buf, pws2->rnd, PWS2_BLOCK_SIZE);
    memset (buf+PWS2_BLOCK_SIZE, 0x00, 2);
    memcpy (buf+PWS2_BLOCK_SIZE+2, pws2->p, strlen (pws2->p));

    gcry_md_hash_buffer (GCRY_MD_SHA1, tmp_salt, buf, strlen(pws2->p)+2+PWS2_BLOCK_SIZE);
    
    gcry_cipher_hd_t ctx;
    gcry_error_t err;
    err = gcry_cipher_open (&ctx, GCRY_CIPHER_BLOWFISH, GCRY_CIPHER_MODE_ECB, GCRY_CIPHER_SECURE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
    {
        free_secure (buf);
        free_secure (tmp_rnd);
        free_secure (tmp_salt);
        return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
    }

    err = gcry_cipher_setkey (ctx, tmp_salt, PWS2_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
    {
        free_secure (buf);
        free_secure (tmp_rnd);
        free_secure (tmp_salt);
        gcry_cipher_close (ctx); 
        return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
    }

    // CIPHER(RND)
    int x; for (x = 0; x < 1000; x++)
    {   
        err = gcry_cipher_encrypt (ctx, tmp_rnd, PWS2_BLOCK_SIZE, tmp_rnd, PWS2_BLOCK_SIZE);
        if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        {
            free_secure (buf);
            free_secure (tmp_rnd);
            free_secure (tmp_salt);
            gcry_cipher_close (ctx); 
            return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
        }
    }
    gcry_cipher_close (ctx);
    pws2_swap_bytes (tmp_rnd, PWS2_BLOCK_SIZE);

    // SHA1_init_state_zero(CIPHER(RND)|0x00,0x00)
    gcry_md_hash_buffer (SHA1_INIT_STATE_ZERO, hrnd, tmp_rnd, PWS2_BLOCK_SIZE+2);

    free_secure (buf);
    free_secure (tmp_rnd);
    free_secure (tmp_salt);
    return 1;
}

/**
 * Verifies the passphrase.
 *
 * @param pws   - The pws2_t object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success and 0 if the key is wrong.
 */
int pws2_check_pwd (pws2_t *pws2, pws_error_t *error)
{   
    unsigned char *hrnd = malloc_secure (PWS2_MAC_SIZE*sizeof(char));

    if (!pws2_calc_hrnd (pws2, hrnd, error))
        { free_secure (hrnd); return pws_error (error, PWFILE_WRONG_KEY, STR_ERROR[PWFILE_WRONG_KEY]); }

    if (memcmp(hrnd, pws2->hrnd, PWS2_MAC_SIZE))
        { free_secure (hrnd); return pws_error (error, PWFILE_WRONG_KEY, STR_ERROR[PWFILE_WRONG_KEY]); }

    free_secure (hrnd);
    return 1;
}

/**
 * Calculates the key which is used for encryption.
 *
 * @param pws2 - The pws2_t context.
 */
void pws2_calc_key (pws2_t *pws2)
{
    unsigned char * buf = malloc_secure ((strlen (pws2->p)+PWS2_MAC_SIZE)*sizeof(char));

    memcpy (buf, pws2->p, strlen (pws2->p));
    memcpy (buf+strlen(pws2->p), pws2->salt, PWS2_MAC_SIZE);

    gcry_md_hash_buffer (GCRY_MD_SHA1, pws2->key, buf, strlen(pws2->p)+PWS2_MAC_SIZE);

    free_secure (buf);
}

/**
 * Reads and decrypts the complete data block (all typed fields incl. the 
 * PWS2 header entries.
 *
 * @param pws2  - The pws2_t context.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The PWSv2 FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws2_read_data_block (pws2_t *pws2, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    unsigned char * block = malloc_secure (PWS2_BLOCK_SIZE*sizeof(char));
    unsigned char * ip    = malloc_secure (PWS2_BLOCK_SIZE*sizeof(char));

    struct stat st;
    fstat (fileno(f), &st);

    buf->data = malloc_secure (st.st_size-ftell(f)*sizeof(char));
    buf->data_len = st.st_size-ftell(f);

    int bytes = fread (block, 1, PWS2_BLOCK_SIZE, f);
    if (bytes != PWS2_BLOCK_SIZE && bytes > 0) 
    { 
        free_secure (block); 
        free_secure (ip); 
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    gcry_error_t err;
    gcry_cipher_hd_t ctx;

    err = gcry_cipher_open (&ctx, GCRY_CIPHER_BLOWFISH, GCRY_CIPHER_MODE_ECB, GCRY_CIPHER_SECURE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
    { 
        free_secure (block);
        free_secure (ip); 
        return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
    }

    err = gcry_cipher_setkey (ctx, pws2->key, PWS2_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
    { 
        free_secure (block);
        free_secure (ip); 
        gcry_cipher_close (ctx); 
        return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
    }

    unsigned long pos = 0;
    while (!feof (f) && pos < (buf->data_len-PWS2_BLOCK_SIZE))
    {
        memcpy (ip, block, PWS2_BLOCK_SIZE);

        pws2_swap_bytes (block, PWS2_BLOCK_SIZE);
        err = gcry_cipher_decrypt (ctx, block, PWS2_BLOCK_SIZE, block, PWS2_BLOCK_SIZE);
        if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { 
            free_secure (block);
            free_secure (ip); 
            gcry_cipher_close (ctx); 
            return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
        }
        pws2_swap_bytes (block, PWS2_BLOCK_SIZE);

        pws2_xormem (block, pws2->IP, PWS2_BLOCK_SIZE);
        memcpy (pws2->IP, ip, PWS2_BLOCK_SIZE);
        memcpy (buf->data+pos, block, PWS2_BLOCK_SIZE);
        pos += PWS2_BLOCK_SIZE;

        bytes = fread (block, 1, PWS2_BLOCK_SIZE, f);
        if (bytes != PWS2_BLOCK_SIZE && bytes > 0) 
        { 
            free_secure (block); 
            free_secure (ip); 
            return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
        }
    }

    gcry_cipher_close (ctx);

    free_secure (block); 
    free_secure (ip); 
    return 1;
}

/**
 * Decrypts the pws2 file and stores the decrypted typed field block inside
 * buf.
 *
 * @param pws2  - The pws2_t context.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The PWSv2 FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws2_dec_file (pws2_t *pws2, pws_buffer_t *buf, FILE *f, pws_error_t * error)
{
    int bytes = 0;

    // RND
    bytes = fread (pws2->rnd, 1, PWS2_BLOCK_SIZE, f);
    if (bytes != PWS2_BLOCK_SIZE) 
    { 
        pws2_clear_passphrase (pws2);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }
    
    // H(RND)
    bytes = fread (pws2->hrnd, 1, PWS2_MAC_SIZE, f);
    if (bytes != PWS2_MAC_SIZE) 
    { 
        pws2_clear_passphrase (pws2);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // SALT
    bytes = fread (pws2->salt, 1, PWS2_MAC_SIZE, f);
    if (bytes != PWS2_MAC_SIZE) 
    { 
        pws2_clear_passphrase (pws2);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // IP
    bytes = fread (pws2->IP, 1, PWS2_IP_SIZE, f);
    if (bytes != PWS2_IP_SIZE) 
    {
        pws2_clear_passphrase (pws2);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // calculate H(RND) and verify passphrase
    if (!pws2_check_pwd (pws2, error))
    {
        pws2_clear_passphrase (pws2);
        return 0; 
    }

    // we have got everything we need to start decrypting the data
    pws2_calc_key (pws2);
    if (!pws2_read_data_block (pws2, buf, f, error))
    { 
        pws2_clear_passphrase (pws2);
        return 0; 
    }

    pws2_clear_passphrase (pws2);
    return 1;
}

/**
 * Encrypts buf and writes the whole encrypted data buffer to f.
 *
 * @param pws2  - The pws2_t object.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws2_write_data_block (pws2_t *pws2, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    unsigned char * block = malloc_secure (PWS2_BLOCK_SIZE*sizeof(char));
    unsigned char * ip    = malloc_secure (PWS2_BLOCK_SIZE*sizeof(char));

    gcry_error_t err;
    gcry_cipher_hd_t ctx;

    err = gcry_cipher_open (&ctx, GCRY_CIPHER_BLOWFISH, GCRY_CIPHER_MODE_ECB, GCRY_CIPHER_SECURE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
    {
        free_secure (block);
        free_secure (ip);
        return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
    }

    err = gcry_cipher_setkey (ctx, pws2->key, PWS2_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
    {
        free_secure (block);
        free_secure (ip);
        gcry_cipher_close (ctx); 
        return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
    }

    unsigned char *ptr = buf->data;
    while (ptr != buf->data+buf->data_len)
    {
        memcpy (block, ptr, PWS2_BLOCK_SIZE);
        pws2_xormem (block, pws2->IP, PWS2_BLOCK_SIZE);

        pws2_swap_bytes (block, PWS2_BLOCK_SIZE);
        err = gcry_cipher_encrypt (ctx, block, PWS2_BLOCK_SIZE, block, PWS2_BLOCK_SIZE);
        if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        {
            free_secure (block);
            free_secure (ip);
            gcry_cipher_close (ctx); 
            return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));
        }
        pws2_swap_bytes (block, PWS2_BLOCK_SIZE);

        if (fwrite (block, 1, PWS2_BLOCK_SIZE, f) != PWS2_BLOCK_SIZE) 
            { free_secure (block); free_secure (ip); return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

        memcpy (pws2->IP, block, PWS2_BLOCK_SIZE);
        ptr += PWS2_BLOCK_SIZE;
    }

    gcry_cipher_close (ctx);

    free_secure (block);
    free_secure (ip);
    return 1;
}

/**
 * Enycrypts all data inside the pws_buffer_t object and writes
 * a PWSv2 file using the FILE object f.
 *
 * @param pws2  - The pws2_t object.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws2_enc_file (pws2_t *pws2, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    // write RND
    int bytes = fwrite (pws2->rnd, 1, PWS2_BLOCK_SIZE, f);
    if (bytes != PWS2_BLOCK_SIZE) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write H(RND)
    bytes = fwrite (pws2->hrnd, 1, PWS2_MAC_SIZE, f);
    if ((bytes != PWS2_MAC_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write SALT
    bytes = fwrite (pws2->salt, 1, PWS2_MAC_SIZE, f);
    if ((bytes != PWS2_MAC_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write IP
    bytes = fwrite (pws2->IP, 1, PWS2_BLOCK_SIZE, f);
    if ((bytes != PWS2_BLOCK_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write the encrypted data block
    if (!pws2_write_data_block (pws2, buf, f, error)) { return 0; }

    return 1;
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
