/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pws3.h"

/**
 * Reads the pws v3 header and stores all recognized typed entries inside hdr.
 * ptr must point to the beginning of the unencrypted data block
 * (pws_buffer_t->data).
 *
 * ptr is positioned at the beginning of the next Password-Entry-Block
 * and returned.
 *
 * @param dat - The pws_data_t object used to store the non default preferences.
 * @param ptr - Points to the start of the (decrypted) header block.
 * @param end - Points to the last valid byte inside ptr.
 *
 * @returns   - The start off the next typed field.
 */
unsigned char * pws3_read_header (pws_data_t *dat, unsigned char *ptr, unsigned char *end)
{
    unsigned int len, l;
    unsigned short w1, w2;

    if ((ptr+5) > end) { return (NULL); }
    memcpy (&len, ptr, sizeof (int));
    unsigned int typ = ptr[4];
    if ((ptr+5+len) > end) { return (NULL); }

    unsigned char * nondef;
    while (typ != PWS_HDR_TYP_EOH)
    {
        switch (typ)
        {
            case PWS_HDR_TYP_VERSION: 
                dat->hdr3.version = malloc (len*sizeof(char));
                dat->hdr3.version_len = len;
                memcpy (dat->hdr3.version, ptr+5, len);
                break;
            case PWS_HDR_TYP_UUID: 
                dat->hdr3.uuid = malloc (len*sizeof(char));
                dat->hdr3.uuid_len = len;
                memcpy (dat->hdr3.uuid, ptr+5, len);
                if (len > (sizeof(int)+(2*sizeof(short))))
                {
                    memcpy (&l,  dat->hdr3.uuid,   sizeof(int));
                    memcpy (&w1, dat->hdr3.uuid+4, sizeof(short));
                    memcpy (&w2, dat->hdr3.uuid+6, sizeof(short));
                    l = ntohl (l); w1 = ntohs (w1); w2 = ntohs (w2);
                    memcpy (dat->hdr3.uuid,    &l, sizeof(int));
                    memcpy (dat->hdr3.uuid+4, &w1, sizeof(short));
                    memcpy (dat->hdr3.uuid+6, &w2, sizeof(short));
                }
                break;
            case PWS_HDR_TYP_NONDEF: 
                nondef = malloc (len*sizeof(char)+1);
                memcpy (nondef, ptr+5, len);
                nondef[len] = 0;

                pws_parse_prefs (&dat->prefs, nondef, len+1);
                free (nondef);
                break;
            case PWS_HDR_TYP_TREE_STATUS: 
                dat->hdr3.tree_status = malloc (len*sizeof(char));
                dat->hdr3.tree_status_len = len;
                memcpy (dat->hdr3.tree_status, ptr+5, len);
                break;
            case PWS_HDR_TYP_LAST_MODIFIED: 
                if (len == 8) // backwards compatibility prior passwordsafe 3.0.9
                {
                    char buf[9];
                    memcpy (buf, ptr+5, 8); buf[8] = 0;
                    dat->hdr3.last_modified = strtol(buf, NULL, 16);
                }
                else
                {
                    memcpy ((void*)&dat->hdr3.last_modified, ptr+5, sizeof(time_t)); 
                }
                break;
            case PWS_HDR_TYP_LAST_SAVE_AUTHOR: // deprecated
                if (len > 4)
                {
                    char buf[5];
                    memcpy (buf, ptr+5, 4);
                    buf[4] = 0;

                    int user_len = strtol(buf, NULL, 16);
                    if ((len-user_len-4) > 0)
                    {
                        if (dat->hdr3.last_uid) free (dat->hdr3.last_uid); // maybe set inside PWS_HDR_TYP_LAST_SAVE_UID
                        if (dat->hdr3.last_host) free (dat->hdr3.last_host); // maybe set inside PWS_HDR_TYP_LAST_SAVE_HOST

                        dat->hdr3.last_uid = malloc (user_len*sizeof(char));
                        dat->hdr3.last_uid_len = user_len;
                        memcpy (dat->hdr3.last_uid, ptr+9, user_len);

                        dat->hdr3.last_host = malloc ((len-4-user_len)*sizeof(char));
                        dat->hdr3.last_host_len = len-4-user_len;
                        memcpy (dat->hdr3.last_host, ptr+9, len-4-user_len);
                    }
                }
                break;
            case PWS_HDR_TYP_LAST_SAVE_APP: 
                dat->hdr3.last_app = malloc (len*sizeof(char));
                dat->hdr3.last_app_len = len;
                memcpy (dat->hdr3.last_app, ptr+5, len);
                break;
            case PWS_HDR_TYP_LAST_SAVE_UID: 
                if (dat->hdr3.last_uid) free (dat->hdr3.last_uid); // maybe set inside PWS_HDR_TYP_LAST_SAVE_AUTHOR
                dat->hdr3.last_uid = malloc (len*sizeof(char));
                dat->hdr3.last_uid_len = len;
                memcpy (dat->hdr3.last_uid, ptr+5, len);
                break;
            case PWS_HDR_TYP_LAST_SAVE_HOST: 
                if (dat->hdr3.last_host) free (dat->hdr3.last_host); // maybe set inside PWS_HDR_TYP_LAST_SAVE_AUTHOR
                dat->hdr3.last_host = malloc (len*sizeof(char));
                dat->hdr3.last_host_len = len;
                memcpy (dat->hdr3.last_host, ptr+5, len);
                break;
            case PWS_HDR_TYP_DB_NAME: 
                dat->hdr3.db_name = malloc (len*sizeof(char));
                dat->hdr3.db_name_len = len;
                memcpy (dat->hdr3.db_name, ptr+5, len);
                break;
            case PWS_HDR_TYP_DB_DESC: 
                dat->hdr3.db_desc = malloc (len*sizeof(char));
                dat->hdr3.db_desc_len = len;
                memcpy (dat->hdr3.db_desc, ptr+5, len);
                break;
            default:
                l = len + 5;
                if (l % PWS3_BLOCK_SIZE)
                    { l += PWS3_BLOCK_SIZE-(l % PWS3_BLOCK_SIZE); }

                dat->hdr3.unknown = realloc (dat->hdr3.unknown, (dat->hdr3.unknown_len+l)*sizeof(char));
                memcpy (dat->hdr3.unknown+dat->hdr3.unknown_len, ptr, l);
                dat->hdr3.unknown_len += l;
                break;
        }

        ptr += len+5;
        if ((len+5) % PWS3_BLOCK_SIZE)
            { ptr += (PWS3_BLOCK_SIZE-((len+5) % PWS3_BLOCK_SIZE)); }

        if ((ptr+5) > end) { return (NULL); }
        memcpy (&len, ptr, sizeof (int));
        typ = ptr[4];
        if ((ptr+5+len) > end) { return (NULL); }
    }

    return (ptr+PWS3_BLOCK_SIZE);
}

/**
 * Reads one pws v3 entry block and stores all recognized entries inside entry.
 * ptr should point to the beginning of the next unencrypted entry block.
 *
 * ptr is positioned at the beginning of the next Password-Entry-Block
 * and returned.
 *
 * @param entry - The pws_entry_t object to be filled.
 * @param ptr   - Points to the start of the (decrypted) entry block.
 * @param end   - Points to the last valid byte inside ptr.
 *
 * @returns The start off the next typed field.
 */
unsigned char * pws3_read_entry (pws_entry_t *entry, unsigned char *ptr, unsigned char *end)
{
    unsigned int len, l;
    unsigned short w1, w2;

    if ((ptr+5) > end) { return (NULL); }
    memcpy (&len, ptr, sizeof (int));
    unsigned int typ = ptr[4];
    if ((ptr+5+len) > end) { return (NULL); }

    while (typ != PWS_ENTRY_TYP_EOE)
    {
        unsigned char *pwd_history;
        switch (typ)
        {
            case PWS_ENTRY_TYP_UUID: 
                entry->uuid = malloc_secure (len*sizeof(char));
                entry->uuid_len = len;
                memcpy (entry->uuid, ptr+5, len);
                if (len > (sizeof(int)+(2*sizeof(short))))
                {
                    memcpy (&l,  entry->uuid,   sizeof(int));
                    memcpy (&w1, entry->uuid+4, sizeof(short));
                    memcpy (&w2, entry->uuid+6, sizeof(short));
                    l = ntohl (l); w1 = ntohs (w1); w2 = ntohs (w2);
                    memcpy (entry->uuid,    &l, sizeof(int));
                    memcpy (entry->uuid+4, &w1, sizeof(short));
                    memcpy (entry->uuid+6, &w2, sizeof(short));
                }
                break;
            case PWS_ENTRY_TYP_GROUP: 
                entry->group = malloc_secure (len*sizeof(char));
                entry->group_len = len;
                memcpy (entry->group, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_TITLE: 
                entry->title = malloc_secure (len*sizeof(char));
                entry->title_len = len;
                memcpy (entry->title, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_USERNAME: 
                entry->username = malloc_secure (len*sizeof(char));
                entry->username_len = len;
                memcpy (entry->username, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_NOTES: 
                entry->notes = malloc_secure (len*sizeof(char));
                entry->notes_len = len;
                memcpy (entry->notes, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_PASSWORD: 
                entry->pwd = malloc_secure (len*sizeof(char));
                entry->pwd_len = len;
                memcpy (entry->pwd, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_CREATION_TIME: 
                memcpy ((void*)&entry->ctime, ptr+5, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_PWD_MODIFICATION_TIME: 
                memcpy ((void*)&entry->pwd_mtime, ptr+5, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_LAST_ACCESS_TIME: 
                memcpy ((void*)&entry->atime, ptr+5, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_PASSWORD_LIFETIME: 
                memcpy ((void*)&entry->ltime, ptr+5, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_PASSWORD_POLICY: 
                entry->pwd_policy = malloc_secure (len*sizeof(char));
                entry->pwd_policy_len = len;
                memcpy (entry->pwd_policy, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_LAST_MODIFICATION_TIME: 
                memcpy ((void*)&entry->mtime, ptr+5, sizeof(time_t));
                break;
            case PWS_ENTRY_TYP_URL: 
                entry->url = malloc_secure (len*sizeof(char));
                entry->url_len = len;
                memcpy (entry->url, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_AUTOTYPE: 
                entry->autotype = malloc_secure (len*sizeof(char));
                entry->autotype_len = len;
                memcpy (entry->autotype, ptr+5, len);
                break;
            case PWS_ENTRY_TYP_PASSWORD_HISTORY: 
                pwd_history = malloc_secure (len*sizeof(char));
                memcpy (pwd_history, ptr+5, len);
                pws_parse_pwd_history (entry, pwd_history, len);
                free_secure (pwd_history);
                break;
            default:
                l = len + 5;
                if (l % PWS3_BLOCK_SIZE)
                    { l += PWS3_BLOCK_SIZE-(l % PWS3_BLOCK_SIZE); }

                entry->unknown = realloc (entry->unknown, (entry->unknown_len+l)*sizeof(char));
                memcpy (entry->unknown+entry->unknown_len, ptr, l);
                entry->unknown_len += l;
                break;
        }

        ptr += len+5;
        if ((len+5) % PWS3_BLOCK_SIZE)
            { ptr += (PWS3_BLOCK_SIZE-((len+5) % PWS3_BLOCK_SIZE)); }

        if ((ptr+5) > end) { return (NULL); }
        memcpy (&len, ptr, sizeof (int));
        typ = ptr[4];
        if ((ptr+5+len) > end) { return (NULL); }
    }

    return (ptr+PWS3_BLOCK_SIZE);
}

/**
 * Decrypts the pws v3 file f. The header data and all 
 * entries are stored inside dat.
 *
 * pws3_init_data() must be called before using this function.
 *
 * @param dat - The pws_data_t object to be filled.
 * @param ctx - The pws_context_t context object.
 * @param f   - The PWSv3 FILE object.
 *
 * @returns 1 on success, otherwise 1 and error->error will be set to the corresponding error code.
 */
int pws3_decrypt (pws_data_t *dat, pws_context_t *ctx, FILE *f)
{
    pws_buffer_t buf;
    buf.data     = NULL;
    buf.data_len = 0;

    if (!pws3_dec_file (&ctx->pws3, &buf, f, &ctx->error))
    {
        free_secure (buf.data);
        return 0;
    }

    unsigned char * ptr = pws3_read_header (dat, buf.data, buf.data+buf.data_len);
    if (!ptr) 
    { 
        ctx->error.error = PWFILE_CORRUPT;
        ctx->error.str_error = strdup(STR_ERROR[PWFILE_CORRUPT]);

        free_secure (buf.data);
        return 0;
    }

    while (ptr != (buf.data+buf.data_len))
    {
        dat->num_entries++;
        dat->entries = realloc (dat->entries, dat->num_entries*sizeof(pws_entry_t*));
        dat->entries[dat->num_entries-1] = malloc (sizeof(pws_entry_t));
        pws_init_entry (dat->entries[dat->num_entries-1]);

        ptr = pws3_read_entry (dat->entries[dat->num_entries-1], ptr, buf.data+buf.data_len);
        if (!ptr) 
        { 
            ctx->error.error = PWFILE_CORRUPT;
            ctx->error.str_error = strdup(STR_ERROR[PWFILE_CORRUPT]);

            free_secure (buf.data);
            return 0;
        }
    }

    free_secure (buf.data);
    return 1;
}

/**
 * Writes one typed field to the buffer ptr.
 *
 * @param ptr      - The buffer to hold the typed entry.
 * @param data     - The data to be written.
 * @param data_len - The size of data.
 * @param typ      - The type of the typed entry.
 */
void pws3_write_typed_field (unsigned char *ptr, void *data, unsigned int data_len, unsigned char typ)
{
    memcpy (ptr, (void*)&data_len, 4);
    memcpy (ptr+4, &typ, 1);
    memcpy (ptr+5, data, data_len);
}

/**
 * Calculate how many bytes will be needed to store dat inside a pws_buffer_t object.
 *
 * @param dat - The password entries.
 *
 * @returns The size in bytes that will be consumed by dat if stored inside pws_buffer_t.
 */
unsigned long pws3_calc_buf_size (pws_data_t *dat)
{
    unsigned long siz = 0;

    // PWSV3 Header Data - always 2 bytes
    dat->hdr3.version_len = 2;

    // typed field: version
    siz += dat->hdr3.version_len + 5;
    if ((dat->hdr3.version_len+5)%PWS3_BLOCK_SIZE)
        siz += PWS3_BLOCK_SIZE-((dat->hdr3.version_len+5)%PWS3_BLOCK_SIZE);

    // typed field: uuid
    if (dat->hdr3.uuid)
    {
        siz += dat->hdr3.uuid_len + 5;
        if ((dat->hdr3.uuid_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.uuid_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: nondef
    char * nondef = pws_nondef_string (&dat->prefs);
    if (nondef)
    {
        siz += strlen(nondef) + 5;
        if ((strlen(nondef)+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((strlen(nondef)+5)%PWS3_BLOCK_SIZE);
        free (nondef);
    }

    // typed field: tree_status
    if (dat->hdr3.tree_status)
    {
        siz += dat->hdr3.tree_status_len + 5;
        if ((dat->hdr3.tree_status_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.tree_status_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_save_author (deprecated, just for backwards compatibility)
    if (dat->hdr3.last_uid && dat->hdr3.last_host)
    {
        unsigned int len = dat->hdr3.last_uid_len + dat->hdr3.last_host_len + 9;
        if (len%PWS3_BLOCK_SIZE)
            len += PWS3_BLOCK_SIZE-(len%PWS3_BLOCK_SIZE);

        siz += len;
    }

    // typed field: last_app
    if (dat->hdr3.last_app)
    {
        siz += dat->hdr3.last_app_len + 5;
        if ((dat->hdr3.last_app_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.last_app_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_uid
    if (dat->hdr3.last_uid)
    {
        siz += dat->hdr3.last_uid_len + 5;
        if ((dat->hdr3.last_uid_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.last_uid_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_host
    if (dat->hdr3.last_host)
    {
        siz += dat->hdr3.last_host_len + 5;
        if ((dat->hdr3.last_host_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.last_host_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: db_name
    if (dat->hdr3.db_name)
    {
        siz += dat->hdr3.db_name_len + 5;
        if ((dat->hdr3.db_name_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.db_name_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: db_desc
    if (dat->hdr3.db_desc)
    {
        siz += dat->hdr3.db_desc_len + 5;
        if ((dat->hdr3.db_desc_len+5)%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-((dat->hdr3.db_desc_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_modified
    if (dat->hdr3.last_modified)
    {
        siz += sizeof(time_t) + 5;
        siz += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // unknown fields
    if (dat->hdr3.unknown)
    {
        siz += dat->hdr3.unknown_len;
    }

    // typed field: End of Header
    siz += PWS3_BLOCK_SIZE;

    int x; for (x = 0; x < dat->num_entries; x++)
    {
        // typed field: uuid
        if (dat->entries[x]->uuid)
        {
            siz += dat->entries[x]->uuid_len + 5;
            if ((dat->entries[x]->uuid_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->uuid_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: group
        if (dat->entries[x]->group)
        {
            siz += dat->entries[x]->group_len + 5;
            if ((dat->entries[x]->group_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->group_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: title
        if (dat->entries[x]->title)
        {
            siz += dat->entries[x]->title_len + 5;
            if ((dat->entries[x]->title_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->title_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: username
        if (dat->entries[x]->username)
        {
            siz += dat->entries[x]->username_len + 5;
            if ((dat->entries[x]->username_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->username_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: notes
        if (dat->entries[x]->notes)
        {
            siz += dat->entries[x]->notes_len + 5;
            if ((dat->entries[x]->notes_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->notes_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: pwd
        if (dat->entries[x]->pwd)
        {
            siz += dat->entries[x]->pwd_len + 5;
            if ((dat->entries[x]->pwd_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->pwd_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: policy
        if (dat->entries[x]->pwd_policy)
        {
            siz += dat->entries[x]->pwd_policy_len + 5;
            if ((dat->entries[x]->pwd_policy_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->pwd_policy_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: url
        if (dat->entries[x]->url)
        {
            siz += dat->entries[x]->url_len + 5;
            if ((dat->entries[x]->url_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->url_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: autotype
        if (dat->entries[x]->autotype)
        {
            siz += dat->entries[x]->autotype_len + 5;
            if ((dat->entries[x]->autotype_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((dat->entries[x]->autotype_len+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: pwd_history
        if (dat->entries[x]->pwd_history || dat->entries[x]->def_pwd_max != -1 || dat->entries[x]->save_pwd_history != -1)
        {
            char *pwd_history = NULL;
            int pwd_history_len = 0;
            pws_pwd_history (&dat->prefs, dat->entries[x], &pwd_history, &pwd_history_len);

            siz += pwd_history_len + 5;
            if ((pwd_history_len+5)%PWS3_BLOCK_SIZE)
                siz += PWS3_BLOCK_SIZE-((pwd_history_len+5)%PWS3_BLOCK_SIZE);

            free_secure (pwd_history);
        }

        // typed field: ctime
        if (dat->entries[x]->ctime)
        {
            siz += sizeof(time_t) + 5;
            siz += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: pwd_mtime
        if (dat->entries[x]->pwd_mtime)
        {
            siz += sizeof(time_t) + 5;
            siz += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: mtime
        if (dat->entries[x]->mtime)
        {
            siz += sizeof(time_t) + 5;
            siz += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: atime
        if (dat->entries[x]->atime)
        {
            siz += sizeof(time_t) + 5;
            siz += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
        }

        // typed field: ltime
        if (dat->entries[x]->ltime)
        {
            siz += sizeof(time_t) + 5;
            siz += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
        }

        // unknown fields
        if (dat->entries[x]->unknown)
        {
            siz += dat->entries[x]->unknown_len;
        }

        // typed field: End of PWS_ENTRY
        siz += PWS3_BLOCK_SIZE;
    }

    return siz;
}

/**
 * Write the Header Block to buf->data.
 *
 * @param dat - The pws_data_t object used to store the non default preferences.
 * @param buf - The pws_buffer_t object.
 *
 * @returns The new position in buf.
 */
unsigned long pws3_write_header (pws_data_t *dat, pws_buffer_t *buf)
{
    unsigned long pos = 0;

    // maybe this header was already filled by pws3_read_header
    free (dat->hdr3.version);

    // PWSV3 Header Data
    dat->hdr3.version     = malloc (2*sizeof(char));
    dat->hdr3.version[0]  = PWSV3_VER & 0xff;
    dat->hdr3.version[1]  = (PWSV3_VER & 0xff00) >> 8;
    dat->hdr3.version_len = 2;

    // typed field: version
    if (dat->hdr3.version)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.version, dat->hdr3.version_len, PWS_HDR_TYP_VERSION);

        pos += dat->hdr3.version_len + 5;
        if ((dat->hdr3.version_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.version_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: uuid
    if (dat->hdr3.uuid)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.uuid, dat->hdr3.uuid_len, PWS_HDR_TYP_UUID);

        pos += dat->hdr3.uuid_len + 5;
        if ((dat->hdr3.uuid_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.uuid_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: nondef
    char * nondef = pws_nondef_string (&dat->prefs);
    if (nondef)
    {
        pws3_write_typed_field (buf->data+pos, nondef, strlen(nondef), PWS_HDR_TYP_NONDEF);

        pos += strlen(nondef) + 5;
        if ((strlen(nondef)+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((strlen(nondef)+5)%PWS3_BLOCK_SIZE);

        free (nondef);
    }

    // typed field: tree_status
    if (dat->hdr3.tree_status)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.tree_status, dat->hdr3.tree_status_len, PWS_HDR_TYP_TREE_STATUS);

        pos += dat->hdr3.tree_status_len + 5;
        if ((dat->hdr3.tree_status_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.tree_status_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_save_author (deprecated, just for backwards compatibility)
    if (dat->hdr3.last_uid && dat->hdr3.last_host)
    {
        unsigned int siz = dat->hdr3.last_uid_len + dat->hdr3.last_host_len + 9;
        char *tmp = malloc ((siz+1)*sizeof(char));
        sprintf (tmp, "%04x", dat->hdr3.last_uid_len);

        memcpy (tmp+4, dat->hdr3.last_uid, dat->hdr3.last_uid_len);
        memcpy (tmp+4+dat->hdr3.last_uid_len, dat->hdr3.last_host, dat->hdr3.last_host_len);

        pws3_write_typed_field (buf->data+pos, tmp, siz-5, PWS_HDR_TYP_LAST_SAVE_AUTHOR);

        if (siz%PWS3_BLOCK_SIZE)
            siz += PWS3_BLOCK_SIZE-(siz%PWS3_BLOCK_SIZE);

        free (tmp);
        pos += siz;
    }

    // typed field: last_app
    if (dat->hdr3.last_app)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.last_app, dat->hdr3.last_app_len, PWS_HDR_TYP_LAST_SAVE_APP);

        pos += dat->hdr3.last_app_len + 5;
        if ((dat->hdr3.last_app_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.last_app_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_uid
    if (dat->hdr3.last_uid)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.last_uid, dat->hdr3.last_uid_len, PWS_HDR_TYP_LAST_SAVE_UID);

        pos += dat->hdr3.last_uid_len + 5;
        if ((dat->hdr3.last_uid_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.last_uid_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_host
    if (dat->hdr3.last_host)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.last_host, dat->hdr3.last_host_len, PWS_HDR_TYP_LAST_SAVE_HOST);

        pos += dat->hdr3.last_host_len + 5;
        if ((dat->hdr3.last_host_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.last_host_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: db_name
    if (dat->hdr3.db_name)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.db_name, dat->hdr3.db_name_len, PWS_HDR_TYP_DB_NAME);

        pos += dat->hdr3.db_name_len + 5;
        if ((dat->hdr3.db_name_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.db_name_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: db_desc
    if (dat->hdr3.db_desc)
    {
        pws3_write_typed_field (buf->data+pos, dat->hdr3.db_desc, dat->hdr3.db_desc_len, PWS_HDR_TYP_DB_DESC);

        pos += dat->hdr3.db_desc_len + 5;
        if ((dat->hdr3.db_desc_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((dat->hdr3.db_desc_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: last_modified
    if (dat->hdr3.last_modified)
    {
        pws3_write_typed_field (buf->data+pos, &dat->hdr3.last_modified, sizeof(time_t), PWS_HDR_TYP_LAST_MODIFIED);

        pos += sizeof(time_t) + 5;
        pos += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // unknown fields
    if (dat->hdr3.unknown)
    {
        memcpy (buf->data+pos, dat->hdr3.unknown, dat->hdr3.unknown_len);

        pos += dat->hdr3.unknown_len;
    }

    // typed field: End of Header
    pws3_write_typed_field (buf->data+pos, NULL, 0, PWS_HDR_TYP_EOH);
    pos += PWS3_BLOCK_SIZE;

    return (pos);
}

/**
 * Write one entry Block to the end of buf->data.
 *
 * @param dat   - The pws_data_t object used to store the non default preferences.
 * @param buf   - The pws_buffer_t object.
 * @param entry - The entry that should be written.
 * @param pos   - the position in buf to store the record
 *
 * @returns The new position in buf.
 */
unsigned long pws3_write_entry (pws_data_t *dat, pws_buffer_t *buf, pws_entry_t *entry, unsigned long pos)
{
    // typed field: uuid
    if (entry->uuid)
    {
        pws3_write_typed_field (buf->data+pos, entry->uuid, entry->uuid_len, PWS_ENTRY_TYP_UUID);

        pos += entry->uuid_len + 5;
        if ((entry->uuid_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->uuid_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: group
    if (entry->group)
    {
        pws3_write_typed_field (buf->data+pos, entry->group, entry->group_len, PWS_ENTRY_TYP_GROUP);

        pos += entry->group_len + 5;
        if ((entry->group_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->group_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: title
    if (entry->title)
    {
        pws3_write_typed_field (buf->data+pos, entry->title, entry->title_len, PWS_ENTRY_TYP_TITLE);

        pos += entry->title_len + 5;
        if ((entry->title_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->title_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: username
    if (entry->username)
    {
        pws3_write_typed_field (buf->data+pos, entry->username, entry->username_len, PWS_ENTRY_TYP_USERNAME);

        pos += entry->username_len + 5;
        if ((entry->username_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->username_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: notes
    if (entry->notes)
    {
        pws3_write_typed_field (buf->data+pos, entry->notes, entry->notes_len, PWS_ENTRY_TYP_NOTES);

        pos += entry->notes_len + 5;
        if ((entry->notes_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->notes_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: pwd
    if (entry->pwd)
    {
        pws3_write_typed_field (buf->data+pos, entry->pwd, entry->pwd_len, PWS_ENTRY_TYP_PASSWORD);

        pos += entry->pwd_len + 5;
        if ((entry->pwd_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->pwd_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: policy
    if (entry->pwd_policy)
    {
        pws3_write_typed_field (buf->data+pos, entry->pwd_policy, entry->pwd_policy_len, PWS_ENTRY_TYP_PASSWORD_POLICY);

        pos += entry->pwd_policy_len + 5;
        if ((entry->pwd_policy_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->pwd_policy_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: url
    if (entry->url)
    {
        pws3_write_typed_field (buf->data+pos, entry->url, entry->url_len, PWS_ENTRY_TYP_URL);

        pos += entry->url_len + 5;
        if ((entry->url_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->url_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: autotype
    if (entry->autotype)
    {
        pws3_write_typed_field (buf->data+pos, entry->autotype, entry->autotype_len, PWS_ENTRY_TYP_AUTOTYPE);

        pos += entry->autotype_len + 5;
        if ((entry->autotype_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((entry->autotype_len+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: pwd_history
    if (entry->pwd_history || entry->def_pwd_max != -1 || entry->save_pwd_history != -1)
    {
        char *pwd_history = NULL;
        int pwd_history_len = 0;
        pws_pwd_history (&dat->prefs, entry, &pwd_history, &pwd_history_len);

        pws3_write_typed_field (buf->data+pos, pwd_history, pwd_history_len, PWS_ENTRY_TYP_PASSWORD_HISTORY);

        pos += pwd_history_len + 5;
        if ((pwd_history_len+5)%PWS3_BLOCK_SIZE)
            pos += PWS3_BLOCK_SIZE-((pwd_history_len+5)%PWS3_BLOCK_SIZE);

        free_secure (pwd_history);
    }

    // typed field: ctime
    if (entry->ctime)
    {
        pws3_write_typed_field (buf->data+pos, &entry->ctime, sizeof(time_t), PWS_ENTRY_TYP_CREATION_TIME);

        pos += sizeof(time_t) + 5;
        pos += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: pwd_mtime
    if (entry->pwd_mtime)
    {
        pws3_write_typed_field (buf->data+pos, &entry->pwd_mtime, sizeof(time_t), PWS_ENTRY_TYP_PWD_MODIFICATION_TIME);

        pos += sizeof(time_t) + 5;
        pos += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: mtime
    if (entry->mtime)
    {
        pws3_write_typed_field (buf->data+pos, &entry->mtime, sizeof(time_t), PWS_ENTRY_TYP_LAST_MODIFICATION_TIME);

        pos += sizeof(time_t) + 5;
        pos += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: atime
    if (entry->atime)
    {
        pws3_write_typed_field (buf->data+pos, &entry->atime, sizeof(time_t), PWS_ENTRY_TYP_LAST_ACCESS_TIME);

        pos += sizeof(time_t) + 5;
        pos += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // typed field: ltime
    if (entry->ltime)
    {
        pws3_write_typed_field (buf->data+pos, &entry->ltime, sizeof(time_t), PWS_ENTRY_TYP_PASSWORD_LIFETIME);

        pos += sizeof(time_t) + 5;
        pos += PWS3_BLOCK_SIZE-((sizeof(time_t)+5)%PWS3_BLOCK_SIZE);
    }

    // unknown fields
    if (entry->unknown)
    {
        memcpy (buf->data+pos, entry->unknown, entry->unknown_len);

        pos += entry->unknown_len;
    }

    // typed field: End of PWS_ENTRY
    pws3_write_typed_field (buf->data+pos, NULL, 0, PWS_ENTRY_TYP_EOE);
    pos += PWS3_BLOCK_SIZE;

    return pos;
}

/**
 * Fills buf->data with all entries and the header inside dat.
 *
 * @param buf - The pws_buffer_t object that should be filled.
 * @param dat - The header and the entries that should be written to pws->data.
 */
void pws3_write_data_buffer (pws_buffer_t *buf, pws_data_t *dat)
{
    buf->data_len = pws3_calc_buf_size (dat);
    buf->data = malloc_secure (buf->data_len*sizeof(char));

    unsigned long pos = pws3_write_header (dat, buf);
    int x; for (x = 0; x < dat->num_entries; x++)
    {
        pos = pws3_write_entry (dat, buf,  dat->entries[x], pos);
    }
}

/**
 * Encrypts all data inside dat to the file f.
 *
 * @param dat - The pws_data_t object, holding all passowrds/notes/header information etc.
 * @param ctx - The pws_context_t context object.
 * @param f   - The FILE object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_encrypt (pws_data_t *dat, pws_context_t *ctx, FILE *f)
{
    pws_buffer_t buf;
    buf.data     = NULL;
    buf.data_len = 0;

    pws3_write_data_buffer (&buf, dat);

    if (!pws3_enc_file (&ctx->pws3, &buf, f, &ctx->error)) 
    {
        free_secure (buf.data);
        return 0;
    }

    free_secure (buf.data);
    return 1;
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
