/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pws3crypt.h"

/**
 * Clears the passphrase.
 *
 * @param ctx - The initialized pws3_t object.
 */
void pws3_clear_passphrase (pws3_t *ctx)
{
    free_secure (ctx->p);
    ctx->p = NULL;
}

/**
 * Initializes the pws3_t context,
 * Must be called before using any other
 * decrypting functions.
 *
 * @param dat - The pws3_t context object containing the context that should be initialized.
 * @param p   - The passphrase.
 * @param enc_only - If set to 1, only the pwd hash will be hold in memory, the pwd itself
 *                   will be cleaned immediately. Only valid if this context will be used
 *                   only for encrypting purposes. If this context is used for decryption also,
 *                   dec_only must be set to 0.
 * @param error - The pws_error_t object.
 */
void pws3_init_ctx (pws3_t *ctx, char *p, int enc_only, pws_error_t *error)
{
    ctx->ptag      = malloc_secure (PWS3_MAC_SIZE*sizeof(char));
    ctx->K         = malloc_secure (PWS3_MAC_SIZE*sizeof(char));
    ctx->L         = malloc_secure (PWS3_MAC_SIZE*sizeof(char));
    ctx->salt      = malloc (PWS3_MAC_SIZE*sizeof(char));
    ctx->IV        = malloc (PWS3_IV_SIZE *sizeof(char));
    ctx->HMAC      = malloc (PWS3_MAC_SIZE*sizeof(char));
    ctx->hptag     = malloc (PWS3_MAC_SIZE*sizeof(char));
    ctx->cbc_ctx   = NULL;
    ctx->p         = NULL;
    ctx->iter      = 0;

    gcry_create_nonce (ctx->K,    PWS3_MAC_SIZE*sizeof(char));
    gcry_create_nonce (ctx->L,    PWS3_MAC_SIZE*sizeof(char));
    gcry_create_nonce (ctx->salt, PWS3_MAC_SIZE*sizeof(char));
    gcry_create_nonce (ctx->IV,   PWS3_IV_SIZE*sizeof(char));

    // clone p
    ctx->p = strdup_secure (p);

    // number of iterations
    ctx->iter = PWS3_MIN_ITERATIONS;

    // calculate H(P')
    pws3_stretch_pwd (ctx->p, ctx->salt, ctx->iter, ctx->ptag);
    pws3_calc_hptag (ctx->hptag, ctx->ptag);

    if (enc_only)
    {
        pws3_clear_passphrase (ctx);
    }
}

/**
 * Frees/trashes all ressources used by the pws3_t context.
 *
 * @param ctx - The pws3_t context object containing the context that should be freed.
 */
void pws3_release_ctx (pws3_t *ctx)
{
    free_secure (ctx->ptag);
    free_secure (ctx->K);
    free_secure (ctx->L);
    free_secure (ctx->p);
    free (ctx->HMAC);
    free (ctx->hptag);
    free (ctx->salt);
    free (ctx->IV);

    ctx->iter  = 0;

    if (ctx->cbc_ctx)
        gcry_cipher_close (*ctx->cbc_ctx);
    free (ctx->cbc_ctx);
}

/**
 * Simply stretch the passphrase p using the salt s and iter
 * iterations to get P'.
 *
 * P' will be stored inside ptag.
 *
 * @param p - The passphrase.
 * @param s - The salt bytes.
 * @param iter - Number of iterations for hashing P'.
 * @param ptag - P' will be stored in ptag.
 */
void pws3_stretch_pwd (char *p, unsigned char *s, unsigned int iter, unsigned char *ptag)
{
    unsigned char *buf = malloc_secure ((strlen(p)+PWS3_MAC_SIZE)*(sizeof(char)));
    strcpy ((char*)buf, p);
    memcpy (buf+strlen(p), s, PWS3_MAC_SIZE);

    // calculate P'
    gcry_md_hash_buffer (GCRY_MD_SHA256, buf, buf, strlen(p)+PWS3_MAC_SIZE);

    unsigned int i;
    for (i = 0; i < iter; i++)
        gcry_md_hash_buffer (GCRY_MD_SHA256, buf, buf, PWS3_MAC_SIZE);

    memcpy (ptag, buf, PWS3_MAC_SIZE);
    free_secure (buf);
}

/**
 * Sets up the gcrypt cbc context pws->cbc_ctx. pws3->K and pws3->L needs
 * to be set before using this function.
 *
 * @param pws3  - The pws3_t context.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_setup_cbc_mode (pws3_t *pws3, pws_error_t *error)
{
    gcry_error_t err;
    // ensure the pws3->cbc_ctx is closed
    if (pws3->cbc_ctx)
    {
        gcry_cipher_close (*pws3->cbc_ctx);
        free (pws3->cbc_ctx);
    }

    // setup cbc mode
    pws3->cbc_ctx = malloc (sizeof(gcry_cipher_hd_t));

    err = gcry_cipher_open (pws3->cbc_ctx, GCRY_CIPHER_TWOFISH, GCRY_CIPHER_MODE_CBC, GCRY_CIPHER_SECURE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_cipher_setkey (*pws3->cbc_ctx, pws3->K, PWS3_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_cipher_setiv  (*pws3->cbc_ctx, pws3->IV, PWS3_IV_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    return 1;
}

/**
 * Decrypts one Twofish CBC block. setup_cbc_mode() needs to be called before
 * using this function.
 *
 * The decrypted Block will be stored inside block.
 *
 * @param pws3  - The pws3_t context.
 * @param block - The block that should be decrypted. The decrypted data will
 *                be stored inside block.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_dec_block_cbc (pws3_t *pws3, unsigned char *block, pws_error_t *error)
{
    gcry_error_t err = gcry_cipher_decrypt (*pws3->cbc_ctx, block, PWS3_BLOCK_SIZE, block, PWS3_BLOCK_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    return 1;
}

/**
 * Encrypte one Twofish CBC block. setup_cbc_mode() needs to be called before
 * using this function.
 *
 * The encrypted Block will be stored inside block.
 *
 * @param pws3  - The pws3_t context.
 * @param block - The block that should be encrypted. The encrypted data will
 *                be stored inside block.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_enc_block_cbc (pws3_t *pws3, unsigned char *block, pws_error_t *error)
{
    gcry_error_t err = gcry_cipher_encrypt (*pws3->cbc_ctx, block, PWS3_BLOCK_SIZE, block, PWS3_BLOCK_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    return 1;
}

/**
 * Decrypts one Twofish EBC block.
 *
 * The decrypted Block will be stored inside block.
 *
 * @param pws3  - The pws3_t context.
 * @param block - The block that should be decrypted. The decrypted data will
 *                be stored inside block.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_dec_block_ecb (pws3_t *pws3, unsigned char *block, pws_error_t *error)
{
    gcry_error_t err;
    gcry_cipher_hd_t ctx;

    err = gcry_cipher_open (&ctx, GCRY_CIPHER_TWOFISH, GCRY_CIPHER_MODE_ECB, GCRY_CIPHER_SECURE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_cipher_setkey (ctx, pws3->ptag, PWS3_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { gcry_cipher_close (ctx); return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_cipher_decrypt (ctx, block, PWS3_BLOCK_SIZE, block, PWS3_BLOCK_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { gcry_cipher_close (ctx); return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    gcry_cipher_close (ctx);

    return 1;
}

/**
 * Encrypts one Twofish EBC block.
 *
 * The encrypted Block will be stored inside block.
 *
 * @param pws3  - The pws3_t context.
 * @param block - The block that should be encrypted. The encrypted data will
 *                be stored inside block.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_enc_block_ecb (pws3_t *pws3, unsigned char *block, pws_error_t *error)
{
    gcry_error_t err;
    gcry_cipher_hd_t ctx;

    err = gcry_cipher_open (&ctx, GCRY_CIPHER_TWOFISH, GCRY_CIPHER_MODE_ECB, GCRY_CIPHER_SECURE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_cipher_setkey (ctx, pws3->ptag, PWS3_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { gcry_cipher_close (ctx); return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_cipher_encrypt (ctx, block, PWS3_BLOCK_SIZE, block, PWS3_BLOCK_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { gcry_cipher_close (ctx); return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    gcry_cipher_close (ctx);

    return 1;
}

/**
 * Calculates SHA256(ptag) and stores the result in hptag.
 *
 * @param hptag - H(P') will be stored in hptag.
 * @param ptag  - P' which will be used to calculate H(P').
 */
void pws3_calc_hptag (unsigned char *hptag, unsigned char *ptag)
{
    // calculate H(P')
    gcry_md_hash_buffer (GCRY_MD_SHA256, hptag, ptag, PWS3_MAC_SIZE);
}

/**
 * Verifies the passphrase. Essentially compares the calculated P' 
 * ptag with H(P') out of the pwsafe file (hptag).
 *
 * @param hptag - H(P')
 * @param ptag  - H(P') will be calculated from ptag and matched
 *                against hptag.
 *
 * @returns 1 on success and 0 if hptag does not matches the calculated H(P').
 */
int pws3_check_pwd (unsigned char *hptag, unsigned char *ptag)
{   
    unsigned char hptag_tmp[PWS3_MAC_SIZE];

    pws3_calc_hptag (hptag_tmp, ptag);

    if (memcmp (hptag, hptag_tmp, PWS3_MAC_SIZE) != 0)
        return 0;
    return 1;
}

/**
 * Reads and decrypts the complete data block (all typed fields incl. The 
 * PWS3 header entries.
 *
 * @param pws3  - The pws3_t context.
 * @param buf   - The pws_buffer_t context.
 * @param f     - The PWSv3 FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_read_data_block (pws3_t *pws3, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    static const char *eof = "PWS3-EOFPWS3-EOF";
    unsigned char * block = malloc_secure (PWS3_BLOCK_SIZE*sizeof(char));

    struct stat st;
    fstat (fileno(f), &st);

    buf->data = malloc_secure ((st.st_size-ftell(f)-PWS3_MAC_SIZE-PWS3_BLOCK_SIZE)*sizeof(char));
    buf->data_len = st.st_size-ftell(f)-PWS3_MAC_SIZE-PWS3_BLOCK_SIZE;

    unsigned long pos = 0;
    while (memcmp (block, eof, PWS3_BLOCK_SIZE) && pos <= buf->data_len)
    {
        if (fread (block, 1, PWS3_BLOCK_SIZE, f) != PWS3_BLOCK_SIZE) 
            { free_secure (block); return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); }

        if (memcmp (block, eof, PWS3_BLOCK_SIZE))
        {
            if (!pws3_dec_block_cbc (pws3, block, error)) { free_secure (block); return 0; }

            memcpy (buf->data+pos, block, PWS3_BLOCK_SIZE);
            pos += PWS3_BLOCK_SIZE;
        }
    }

    if (fread (pws3->HMAC, 1, PWS3_MAC_SIZE, f) != PWS3_MAC_SIZE) 
        { free_secure (block); return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); }

    free_secure (block);
    return 1;
}

/**
 * Calculates SHA256 (DATA) - the checksum at the end of the PWS3 file.
 *
 * @param pws3  - The pws3_t object.
 * @param buf   - The pws_buffer_t object.
 * @param hmac  - The checksum will be stored inside hmac.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_calc_chksum (pws3_t *pws3, pws_buffer_t *buf, unsigned char *hmac, pws_error_t *error)
{
    unsigned char * ptr = buf->data;
    gcry_md_hd_t ctx;
    gcry_error_t err;

    err = gcry_md_open   (&ctx, GCRY_MD_SHA256, GCRY_MD_FLAG_SECURE|GCRY_MD_FLAG_HMAC);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    err = gcry_md_setkey (ctx, pws3->L, PWS3_MAC_SIZE);
    if (gcry_err_code(err) != GPG_ERR_NO_ERROR) 
        { gcry_md_close (ctx); return pws_error (error, PWFILE_GCRYPT_ERROR, gcry_strerror (err));}

    while ((ptr-buf->data) != buf->data_len)
    {
        if ((ptr+5) > (buf->data+buf->data_len)) { gcry_md_close (ctx); return (0); }
        unsigned int len = 0;
        memcpy (&len, ptr, sizeof (int));

        if ((ptr+5+len) > (buf->data+buf->data_len)) { gcry_md_close (ctx); return (0); }
        gcry_md_write (ctx, ptr+5, len);

        ptr += len+5;
        if ((len+5)%PWS3_BLOCK_SIZE)
            ptr += PWS3_BLOCK_SIZE-((len+5)%PWS3_BLOCK_SIZE);
    }

    unsigned char *hmac_tmp = gcry_md_read (ctx, GCRY_MD_SHA256);

    memcpy (hmac, hmac_tmp, PWS3_MAC_SIZE);
    gcry_md_close (ctx);

    return (1);
}

/**
 * Verifies the checksum at the end of the data block.
 *
 * This function cannot be executed before read_data_block() was successfully executed.
 *
 * @param pws3  - The pws3_t context.
 * @param buf   - The pws_buffer_t object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_verify_chksum (pws3_t *pws3, pws_buffer_t *buf, pws_error_t *error)
{
    unsigned char *hmac = malloc_secure(sizeof(char)*PWS3_MAC_SIZE);

    if (!pws3_calc_chksum (pws3, buf, hmac, error) || memcmp (hmac, pws3->HMAC, PWS3_MAC_SIZE))
    {
        free_secure (hmac);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]);
    }

    free_secure (hmac);
    return 1;
}

/**
 * Encrypts buf->data and writes the whole encrypted data buffer to f.
 *
 * @param pws3  - The pws3_t object.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_write_data_block (pws3_t *pws3, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    static const char *eof = "PWS3-EOFPWS3-EOF";

    unsigned char *ptr = buf->data;
    while (ptr != buf->data+buf->data_len)
    {
        if (!pws3_enc_block_cbc (pws3, ptr, error))
            return 0;

        if (fwrite (ptr, 1, PWS3_BLOCK_SIZE, f) != PWS3_BLOCK_SIZE) 
            { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

        ptr += PWS3_BLOCK_SIZE;
    }

    if (fwrite (eof, 1, PWS3_BLOCK_SIZE, f) != PWS3_BLOCK_SIZE) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    return 1;
}

/**
 * Decrypts the PWS3-file and stores the decrypted typed field block inside
 * buf->data.
 *
 * @param pws3  - The pws3_t context.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The PWSv3 FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_dec_file (pws3_t *pws3, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    unsigned char ver[4];
    int bytes = 0;

    // header: "PWS3"
    bytes = fread (ver, 1, 4, f);
    if ((bytes != 4) || (strncmp ((char*)ver,"PWS3",4) != 0)) 
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_WRONG_FORMAT, STR_ERROR[PWFILE_WRONG_FORMAT]); 
    }

    // SALT
    bytes = fread (pws3->salt, 1, PWS3_MAC_SIZE, f);
    if (bytes != PWS3_MAC_SIZE) 
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // number of iterations
    bytes = fread ((void*)&pws3->iter, 1, 4, f);
    if (bytes != 4)
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // H(P')
    bytes = fread (pws3->hptag, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // calculate H(P') and verify passphrase
    pws3_stretch_pwd (pws3->p, pws3->salt, pws3->iter, pws3->ptag);
    if (!pws3_check_pwd (pws3->hptag, pws3->ptag)) 
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_WRONG_KEY, STR_ERROR[PWFILE_WRONG_KEY]); 
    }

    // TWOFISH(B1) and TWOFISH(B2)
    bytes = fread (pws3->K, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // TWOFISH(B3) and TWOFISH(B4)
    bytes = fread (pws3->L, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
    {   
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // IV
    bytes = fread (pws3->IV, 1, PWS3_IV_SIZE, f);
    if ((bytes != PWS3_IV_SIZE)) 
    {
        pws3_clear_passphrase (pws3);
        return pws_error (error, PWFILE_CORRUPT, STR_ERROR[PWFILE_CORRUPT]); 
    }

    // decrypt K
    if (!pws3_dec_block_ecb (pws3, pws3->K, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
    }
    if (!pws3_dec_block_ecb (pws3, pws3->K+PWS3_BLOCK_SIZE, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
    }

    // decrypt L
    if (!pws3_dec_block_ecb (pws3, pws3->L, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
     }
    if (!pws3_dec_block_ecb (pws3, pws3->L+PWS3_BLOCK_SIZE, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
    }

    // we have got everything we need to start decrypting the data
    if (!pws3_setup_cbc_mode (pws3, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
    }

    if (!pws3_read_data_block (pws3, buf, f, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
    }

    // verify the checksum at the end of the pws file
    if (!pws3_verify_chksum (pws3, buf, error)) 
    { 
        pws3_clear_passphrase (pws3); 
        return 0; 
    }

    pws3_clear_passphrase (pws3);
    return 1;
}

/**
 * Enycrypts all data inside the pws_buffer_t object buf and writes
 * a PWSv3 file using the FILE object f. pws3 must be already initialized, that
 * includes the passphrase checksum.
 *
 * @param pws3  - The pws3_t object.
 * @param buf   - The pws_buffer_t object.
 * @param f     - The FILE object.
 * @param error - The pws_error_t object.
 *
 * @returns 1 on success, otherwise 0 and error->error will be set to the corresponding error code.
 */
int pws3_enc_file (pws3_t *pws3, pws_buffer_t *buf, FILE *f, pws_error_t *error)
{
    // calculate HMAC checksum
    pws3_calc_chksum (pws3, buf, pws3->HMAC, error);

    // header: "PWS3"
    int bytes = fwrite ("PWS3", 1, 4, f);
    if (bytes != 4) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // setup symmetric twofish encryption
    pws3_setup_cbc_mode (pws3, error);

    // write SALT
    bytes = fwrite (pws3->salt, 1, PWS3_MAC_SIZE, f);
    if (bytes != PWS3_MAC_SIZE) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    bytes = fwrite ((void*)&pws3->iter, 1, 4, f);
    if (bytes != 4) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write H(P')
    bytes = fwrite (pws3->hptag, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write TWOFISH(B1) and TWOFISH(B2)
    if (!pws3_enc_block_ecb (pws3, pws3->K, error)) { return 0; }
    if (!pws3_enc_block_ecb (pws3, pws3->K+PWS3_BLOCK_SIZE, error)) { return 0; }

    bytes = fwrite (pws3->K, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write TWOFISH(B3) and TWOFISH(B4)
    if (!pws3_enc_block_ecb (pws3, pws3->L, error)) { return 0; }
    if (!pws3_enc_block_ecb (pws3, pws3->L+PWS3_BLOCK_SIZE, error)) { return 0; }

    bytes = fwrite (pws3->L, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write IV
    bytes = fwrite (pws3->IV, 1, PWS3_IV_SIZE, f);
    if ((bytes != PWS3_IV_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    // write the encrypted data block
    if (!pws3_write_data_block (pws3, buf, f, error)) { return 0; }

    // and finally write the HMAC chksum
    bytes = fwrite (pws3->HMAC, 1, PWS3_MAC_SIZE, f);
    if ((bytes != PWS3_MAC_SIZE)) 
        { return pws_error (error, PWFILE_WRITE_ERROR, strerror (errno)); }

    return 1;
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
