/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "secmem.h"

/**
 * Trashes siz bytes at buf.
 *
 * @param buf - The buffer that should be trashed.
 * @param siz - The number of bytes that should be overidden.
 */
void trash_memory (void *buf, size_t siz)
{
    if (buf && siz)
    {
        int x;
        for (x = 0; x < 30; x++)
        {
            memset (buf, 0xff, siz);
            memset (buf, 0xaa, siz);
            memset (buf, 0x55, siz);
            memset (buf, 0x00, siz);
        }
    }
}

/**
 * Simple malloc() wrapper.
 * Memory allocated with malloc_secure() must be freed using free_secure().
 */
void * malloc_secure (size_t siz)
{
    void *tmp = NULL;
    if (siz)
    {
        siz += sizeof(size_t);
        tmp = malloc (siz);
#ifndef HAVE_MLOCKALL
#ifdef HAVE_MLOCK
        static int warn = 0;
        if (mlock (tmp, siz) == -1 && warn == 0)
        {
            fprintf (stderr, "libpws - mlock() failed, using insecure memory\n");
            warn = 1;
        }
#endif
#ifdef MINGW32
        static int warn = 0;
        if (!VirtualLock (tmp, siz) && warn == 0)
        {
            fprintf (stderr, "libpws - VirtualLock() failed, using insecure memory\n");
            warn = 1;
        }
#endif
#endif
        memcpy (tmp, &siz, sizeof(size_t));

        tmp += sizeof(size_t);
    }

    return (tmp);
}

/**
 * Simple realloc() wrapper, that trashs all memory before giving 
 * it back to the system.
 *
 * Memory allocated with realloc_secure() must be freed using free_secure(), resp.
 * reallocated using realloc_secure().
 */
void * realloc_secure (void *buf, size_t newsiz)
{
    size_t oldsiz;
    void *tmp = NULL;

    if (buf && newsiz)
    {
        memcpy (&oldsiz, buf-sizeof(size_t), sizeof(size_t));
        oldsiz -= sizeof(size_t);

        tmp = malloc_secure (newsiz);
        if (oldsiz < newsiz)
            { memcpy (tmp, buf, oldsiz); }
        else
            { memcpy (tmp, buf, newsiz); }

        free_secure (buf);
    }
    else if (newsiz)
        { tmp = malloc_secure (newsiz); }

    return (tmp);
}

/**
 * Simple free() wrapper, that trashs all memory before giving it 
 * back to the system.
 * Can only be used to free memory allocated using malloc_secure(), realloc_secure() or
 * strdup_secure().
 */
void free_secure (void *buf)
{
    size_t siz;
    if (buf)
    {
        memcpy (&siz, buf-sizeof(size_t), sizeof(size_t));
        trash_memory (buf-sizeof(size_t), siz);

        free (buf-sizeof(size_t));
#ifndef HAVE_MLOCKALL
#ifdef HAVE_MLOCK
        munlock (buf-sizeof(size_t), siz);
#endif
#ifdef MINGW32
        VirtualUnlock (buf-sizeof(size_t), siz);
#endif
#endif
    }
}

/**
 * Simple strdup() wrapper.
 *
 * Memory allocated with strdup_secure() must be freed using free_secure(), resp.
 * reallocated using realloc_secure().
 */
char * strdup_secure (char *s)
{
    char * tmp = NULL;
    if (s)
    {
        tmp = malloc_secure (strlen(s)+1);
        strcpy (tmp, s);
    }

    return (tmp);
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
