/***************************************************************************
 *                                                                         *
 *   copyright (C) 2007-2008 by Timo Benk                                  *
 *   email: tbenk@pwsafe.de                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "util.h"

/**
 * Adds entry to the pws object.
 *
 * @param pws  - The pws_data_t object.
 * @param entry - The entry to be added.
 */
void pws_add_entry (pws_data_t *pws, pws_entry_t *entry)
{
    pws->entries = realloc (pws->entries, (pws->num_entries+1)*sizeof(pws_entry_t*));
    pws->num_entries += 1;

    pws->entries[pws->num_entries-1] = entry;
}

/**
 * Deletes entry from the pws object.
 *
 * @param pws  - The pws_data_t object.
 * @param entry - The entry to be added.
 */
void pws3_del_entry (pws_data_t *pws, pws_entry_t *entry)
{
    int found = 0;
    int x = 0; for (x = 0; x < pws->num_entries; x++)
    {
        if (found)
            { pws->entries[x-1] = pws->entries[x]; }

        if (pws->entries[x] == entry)
        { 
            found = 1; 
            pws_release_entry (entry);
            free (entry);
        }
    }

    pws->num_entries -= 1;
    pws->entries = realloc (pws->entries, pws->num_entries*sizeof(pws_entry_t*));
}

/**
 * Allocates a new pws entry object. All empty parameters must be set to NULL.
 *
 * @param uuid            - UUID.
 * @param uuid_len        - Length of uuid.
 * @param group           - Group.
 * @param group_len       - Length of group.
 * @param title           - Title.
 * @param title_len       - Length of title.
 * @param username        - Username.
 * @param username_len    - Length of username.
 * @param notes           - Notes.
 * @param notes_len       - Length of notes.
 * @param pwd             - Password.
 * @param pwd_len         - Length of pwd.
 * @param pwd_policy      - Password policy.
 * @param pwd_policy_len  - Length of pwd_policy.
 * @param url             - URL.
 * @param url_len         - Length of url.
 * @param autotype        - Autotype.
 * @param autotype_len    - Length of autotype.
 * @param save_pwd_history - Whether password history should be maintained.
 * @param def_pwd_max     - Maintain max def_pwd_max historical passwords.
 * @param unknown         - Unknown fields.
 * @param unknown_len     - Unknown fields length.
 * @param ctime           - Creation time.
 * @param pwd_mtime       - Password modification time.
 * @param mtime           - Entry modification time.
 * @param atime           - Last access time.
 * @param ltime           - Password lifetime.
 *
 * @returns the new entry object which must be freed using free_secure.
 */
pws_entry_t * pws_new_entry (unsigned char *uuid, int uuid_len, char *group, int group_len,
                               char *title, int title_len, char *username, int username_len,
                               char *notes, int notes_len, char *pwd, int pwd_len,
                               char *pwd_policy, int pwd_policy_len, char *url, int url_len,
                               char *autotype, int autotype_len, int save_pwd_history, int def_pwd_max,
                               char *unknown, unsigned int unknown_len,
                               time_t ctime, time_t pwd_mtime, time_t mtime, time_t atime, time_t ltime)
{
    pws_entry_t * entry = malloc (sizeof(pws_entry_t));
    pws_init_entry (entry);

    entry->uuid_len       = uuid_len;
    entry->group_len      = group_len;
    entry->title_len      = title_len;
    entry->username_len   = username_len;
    entry->notes_len      = notes_len;
    entry->pwd_len        = pwd_len;
    entry->pwd_policy_len = pwd_policy_len;
    entry->url_len        = url_len;
    entry->autotype_len   = autotype_len;
    entry->unknown_len    = unknown_len;

    entry->save_pwd_history = save_pwd_history;
    entry->def_pwd_max      = def_pwd_max;

    if (entry->uuid_len)
    {
        entry->uuid = malloc_secure (uuid_len*sizeof(char));
        memcpy (entry->uuid, uuid, uuid_len);
    }

    if (entry->group_len)
    {
        entry->group = malloc_secure (group_len*sizeof(char));
        memcpy (entry->group, group, group_len);
    }

    if (entry->title_len)
    {
        entry->title = malloc_secure (title_len*sizeof(char));
        memcpy (entry->title, title, title_len);
    }

    if (entry->url_len)
    {
        entry->url = malloc_secure (url_len*sizeof(char));
        memcpy (entry->url, url, url_len);
    }

    if (entry->autotype_len)
    {
        entry->autotype = malloc_secure (autotype_len*sizeof(char));
        memcpy (entry->autotype, autotype, autotype_len);
    }

    if (entry->username_len)
    {
        entry->username = malloc_secure (username_len*sizeof(char));
        memcpy (entry->username, username, username_len);
    }

    if (entry->notes_len)
    {
        entry->notes = malloc_secure (notes_len*sizeof(char));
        memcpy (entry->notes, notes, notes_len);
    }

    if (entry->pwd_len)
    {
        entry->pwd = malloc_secure (pwd_len*sizeof(char));
        memcpy (entry->pwd, pwd, pwd_len);
    }

    if (entry->pwd_policy_len)
    {
        entry->pwd_policy = malloc_secure (pwd_policy_len*sizeof(char));
        memcpy (entry->pwd_policy, pwd_policy, pwd_policy_len);
    }

    if (entry->unknown_len)
    {
        entry->unknown = malloc_secure (unknown_len*sizeof(char));
        memcpy (entry->unknown, unknown, unknown_len);
    }

    entry->ctime     = ctime;
    entry->pwd_mtime = pwd_mtime;
    entry->mtime     = mtime;
    entry->atime     = atime;
    entry->ltime     = ltime;

    return entry;
}

/**
 * Sets the pws v3  header infomation.
 *
 * @param pws           - The pws object.
 * @param uuid          - The Uuid.
 * @param uuid_len      - Length in bytes of uuid.
 * @param last_app      - Application, that performed last save.
 * @param last_app_len  - Length of last_app.
 * @param last_uid      - Uid, who performed last save.
 * @param last_uid_len  - Length of uid.
 * @param last_host     - Host, that performed last save.
 * @param last_host_len - Length of host.
 * @param db_name       - DB name.
 * @param db_name_len   - Length of db_name.
 * @param db_desc       - DB description.
 * @param db_desc_len   - Length of db_desc.
 * @param unknown       - Unknown header fields.
 * @param unknown_len   - Length of unknown.
 * @param last_modified - Timestamp of last save.
 */
void pws_set_header (pws_data_t *pws, unsigned char *uuid, int uuid_len, char *last_app, int last_app_len,
                      char *last_uid, int last_uid_len, char *last_host, int last_host_len,
                      char *db_name, int db_name_len, char *db_desc, int db_desc_len,
                      char *unknown, unsigned int unknown_len,
                      time_t last_modified)
{
    pws2_init_hdr (&pws->hdr2);
    pws3_init_hdr (&pws->hdr3);

    pws->hdr3.uuid_len      = uuid_len;
    pws->hdr3.last_app_len  = last_app_len;
    pws->hdr3.last_uid_len  = last_uid_len;
    pws->hdr3.last_host_len = last_host_len;
    pws->hdr3.db_name_len   = db_name_len;
    pws->hdr3.db_desc_len   = db_desc_len;
    pws->hdr3.unknown_len   = unknown_len;

    if (pws->hdr3.uuid_len)
    {
        pws->hdr3.uuid = malloc (uuid_len*sizeof(char));
        memcpy (pws->hdr3.uuid, uuid, uuid_len);
    }

    if (pws->hdr3.last_app_len)
    {
        pws->hdr3.last_app = malloc (last_app_len*sizeof(char));
        memcpy (pws->hdr3.last_app, last_app, last_app_len);
    }

    if (pws->hdr3.last_uid_len)
    {
        pws->hdr3.last_uid = malloc (last_uid_len*sizeof(char));
        memcpy (pws->hdr3.last_uid, last_uid, last_uid_len);
    }

    if (pws->hdr3.last_host_len)
    {
        pws->hdr3.last_host = malloc (last_host_len*sizeof(char));
        memcpy (pws->hdr3.last_host, last_host, last_host_len);
    }

    if (pws->hdr3.db_name_len)
    {
        pws->hdr3.db_name = malloc (db_name_len*sizeof(char));
        memcpy (pws->hdr3.db_name, db_name, db_name_len);
    }

    if (pws->hdr3.db_desc_len)
    {
        pws->hdr3.db_desc = malloc (db_desc_len*sizeof(char));
        memcpy (pws->hdr3.db_desc, db_desc, db_desc_len);
    }

    if (pws->hdr3.unknown_len)
    {
        pws->hdr3.unknown = malloc (unknown_len*sizeof(char));
        memcpy (pws->hdr3.unknown, unknown, unknown_len);
    }

    pws->hdr3.last_modified = last_modified;
}

/**
 * Sets the preferences stored inside the passwordsafe file.
 * Only non-default values will be saved.
 *
 * @param pws                          - The pws  object.
 * @param showpwdefault                - Show the password in cleartext by default.
 * @param showpasswordintree           - Show the password in the tree view.
 * @param sortascending                - Sort ascending by default.
 * @param usedefuser                   - Use default user.
 * @param saveimmediately              - Save immediately on any change.
 * @param pwuselowercase               - Password policy: use lower case characters.
 * @param pwuseuppercase               - Password policy: use upper case characters.
 * @param pwusedigits                  - Password policy: use digits.
 * @param pwusesysmbols                - Password policy: use symbols.
 * @param pwusehexdigit                - Password policy: use (only) hex numbers.
 * @param pweasyvision                 - Password policy: use only easy to read characters.
 * @param isutf8                       - Textual data is stored in UTF8 format.
 * @param displayexpandedaddeditdialog - ???
 * @param maintaindatetimestamps       - Maintain timestamps.
 * @param savepasswordhistory          - Save password history (-1 == not set, 0 == do not save history, 1 == save history).
 * @param numpwhistorydefault          - Max number of history entries (-1 == not set, 0 == do not save history, 1 == save history).
 * @param shownotesdefault             - Show notes in cleartext by default.
 * @param showusernameintree           - Show the username in the tree view.
 * @param unknown                      - Unknown preferences.
 * @param unknown_len                  - Length of unknown.
 * @param defaultusername              - Default username.
 * @param defaultusername_len          - Length of defaultusername.
 * @param defaultautootypestring       - Default autotype string.
 * @param defaultautootypestring_len   - Length of defaultautootypestring.
 */
void pws_set_prefs (pws_data_t *pws, int showpwdefault, int showpasswordintree, int sortascending, int usedefuser,
                     int saveimmediately, int pwuselowercase, int pwuseuppercase, int pwusedigits, int pwusesysmbols,
                     int pwusehexdigit, int pweasyvision, int isutf8, int displayexpandedaddeditdialog, int maintaindatetimestamps,
                     int savepasswordhistory, int shownotesdefault, int showusernameintree, 
                     int pwlendefault, int idletimeout, int treedisplaystatusatopen, int numpwhistorydefault, 
                     char *unknown, unsigned int unknown_len,
                     char * defaultusername, int defaultusername_len, char * defaultautootypestring, int defaultautootypestring_len)
{
    pws_init_prefs (&pws->prefs);

    pws->prefs.showpwdefault           = showpwdefault;
    pws->prefs.showpasswordintree      = showpasswordintree;
    pws->prefs.sortascending           = sortascending;
    pws->prefs.usedefuser              = usedefuser;
    pws->prefs.saveimmediately         = saveimmediately;
    pws->prefs.pwuselowercase          = pwuselowercase;
    pws->prefs.pwuseuppercase          = pwuseuppercase;
    pws->prefs.pwusedigits             = pwusedigits;
    pws->prefs.pwusesysmbols           = pwusesysmbols;
    pws->prefs.pwusehexdigit           = pwusehexdigit;
    pws->prefs.pweasyvision            = pweasyvision;
    pws->prefs.isutf8                  = isutf8;
    pws->prefs.displayexpandedaddeditdialog = displayexpandedaddeditdialog;
    pws->prefs.maintaindatetimestamps  = maintaindatetimestamps;
    pws->prefs.savepasswordhistory     = savepasswordhistory;
    pws->prefs.shownotesdefault        = shownotesdefault;
    pws->prefs.showusernameintree      = showusernameintree;
    pws->prefs.pwlendefault            = pwlendefault;
    pws->prefs.treedisplaystatusatopen = treedisplaystatusatopen;
    pws->prefs.numpwhistorydefault     = numpwhistorydefault;

    if (idletimeout > 0)
    {
        pws->prefs.idletimeout = idletimeout;
    }

    pws->prefs.defaultusername_len        = defaultusername_len;
    pws->prefs.defaultautootypestring_len = defaultautootypestring_len;
    pws->prefs.unknown_len                = unknown_len;

    if (pws->prefs.defaultusername_len)
    {
        pws->prefs.defaultusername = malloc (defaultusername_len*sizeof(char));
        memcpy (pws->prefs.defaultusername, defaultusername, defaultusername_len);
    }

    if (pws->prefs.defaultautootypestring_len)
    {
        pws->prefs.defaultautootypestring = malloc (defaultautootypestring_len*sizeof(char));
        memcpy (pws->prefs.defaultautootypestring, defaultautootypestring, defaultautootypestring_len);
    }

    if (pws->prefs.unknown_len)
    {
        pws->prefs.unknown = malloc (unknown_len*sizeof(char));
        memcpy (pws->prefs.unknown, unknown, unknown_len);
    }
}

/**
 * Adds pwd to the password history list of entry.
 *
 * @param entry - The pws_entry_t object.
 * @param pwd   - The historical password.
 * @param timestamp - The creation date of pwd.
 */
void pws_add_hst_entry (pws_entry_t *entry, char *pwd, int pwd_len, time_t timestamp)
{
    entry->pwd_history = realloc (entry->pwd_history, (entry->pwd_history_len+1)*sizeof(pws_pwd_hst_entry_t));
    entry->pwd_history[entry->pwd_history_len].pwd = malloc_secure (pwd_len*sizeof(char));

    memcpy (entry->pwd_history[entry->pwd_history_len].pwd, pwd, pwd_len);
    entry->pwd_history[entry->pwd_history_len].pwd_len = pwd_len;
    entry->pwd_history[entry->pwd_history_len].timestamp = timestamp;

    entry->pwd_history_len += 1;
}

/**
 * Verifies the passphrase hash in ctx->pws2.htag with the
 * given passphras p.
 *
 * @param ctx  - The pws_context_t context object.
 * @param p    - The passphrase to be verified.
 * 
 * @returns 1 on success, otherwise 0.
 */
int pws_verify_pwd (pws_context_t *ctx, char *p)
{
    unsigned char *hrnd = malloc_secure (PWS2_MAC_SIZE*sizeof(char));
    char * tmp_p = ctx->pws2.p;

    ctx->pws2.p = strdup_secure(p);

    // calculate H(RND)
    pws2_calc_hrnd (&ctx->pws2, hrnd, &ctx->error);
    
    int ret = 0;
    if (memcmp (hrnd, ctx->pws2.hrnd, PWS2_MAC_SIZE) == 0) { ret = 1;}

    free_secure(ctx->pws2.p);
    free_secure(hrnd);

    ctx->pws2.p = tmp_p;
    return (ret);
}

/* vim:set expandtab tabstop=4 shiftwidth=4 softtabstop=4 nowrap: */
